"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Eagle(Monster):
    """
    Keen Sight.
      The eagle has advantage on Wisdom (Perception) checks that rely on
      sight.
    Talons.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 4 (1d4 +
      2) slashing damage.
    """
    name = "Eagle"
    description = "Small beast, unaligned"
    challenge_rating = 0
    armor_class = 12
    skills = "Perception +4"
    senses = "Passive Perception 14"
    languages = ""
    strength = Ability(6)
    dexterity = Ability(15)
    constitution = Ability(10)
    intelligence = Ability(2)
    wisdom = Ability(14)
    charisma = Ability(7)
    speed = 10
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 3
    hit_dice = "1d6"
    spells = []


class EarthElemental(Monster):
    """
    Earth Glide.
      The elemental can burrow through nonmagical, unworked earth and stone.
      While doing so, the elemental doesn't disturb the material it moves
      through.
    Siege Monster.
      The elemental deals double damage to objects and structures.
    Multiattack.
      The elemental makes two slam attacks.
    Slam.
      Melee Weapon Attack: +8 to hit, reach 10 ft., one target. Hit: 14 (2d8
      + 5) bludgeoning damage.
    """
    name = "Earth Elemental"
    description = "Large elemental, neutral"
    challenge_rating = 5
    armor_class = 17
    skills = ""
    senses = "Darkvision 60 ft., Tremorsense 60 ft., Passive Perception 10"
    languages = "Terran"
    strength = Ability(20)
    dexterity = Ability(8)
    constitution = Ability(20)
    intelligence = Ability(5)
    wisdom = Ability(10)
    charisma = Ability(5)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 126
    hit_dice = "12d10"
    spells = []

# MTF
class EarthElementalMyrmidon(Monster):
    """Magic Weapons.
      The myrmidon's weapon attacks are magical.
    Multiattack.
      The myrmidon makes two maul attacks.
    Maul.
      *Melee Weapon Attack:* +7 to hit, reach 5 ft., one
      target. *Hit:* 11 (2d6 + 4) bludgeoning damage.
    Thunderous Strike (Recharge 6).
      The myrmidon makes one maul attack. On a hit, the target takes
      an extra 16 (3d10) thunder damage, and the target must succeed
      on a DC 14 Strength saving throw or be knocked prone.

    Elemental myrmidons are elementals conjured and bound by magic
    into ritually created suits of plate armor. In this form, they
    possess no recollection of their former existence as free
    elementals. They exist only to follow the commands of their
    creators.

    """
    name = "Earth Elemental Myrmidon"
    description = "Medium elemental, neutral"
    challenge_rating = 7
    armor_class = 18
    damage_resistances = "bludgeoning, piercing, and slashing from nonmagical weapons"
    damage_immunities = "poison"
    condition_immunities = "paralyzed, petrified, poisoned, prone"
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "Terran, one language of its creator's choice"
    strength = Ability(18)
    dexterity = Ability(10)
    constitution = Ability(17)
    intelligence = Ability(8)
    wisdom = Ability(10)
    charisma = Ability(10)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 127
    hit_dice = "17d8 + 51"


class Efreeti(Monster):
    """
    Elemental Demise.
      If the efreeti dies, its body disintegrates in a flash of fire and
      puff of smoke, leaving behind only equipment the djinni was wearing or
      carrying.
    Innate Spellcasting.
      The efreeti's innate spell casting ability is Charisma (spell save DC
      15, +7 to hit with spell attacks). It can innately cast the following
      spells, requiring no material components:
      
  
      
      At will: detect magic
      
      3/day: enlarge/reduce, tongues
      
      1/day each: conjure elemental (fire elemental only), gaseous form,
      invisibility, major image, plane shift, wall of fire
    Variant: Genie Powers.
      Genies have a variety of magical capabilities, including spells. A few
      have even greater powers that allow them to alter their appearance or
      the nature of reality.
      
  
      
      Disguises.
      
      Some genies can veil themselves in illusion to pass as other similarly
      shaped creatures. Such genies can innately cast the disguise self
      spell at will, often with a longer duration than is normal for that
      spell. Mightier genies can cast the true polymorph spell one to three
      times per day, possibly with a longer duration than normal. Such
      genies can change only their own shape, but a rare few can use the
      spell on other creatures and objects as well.
      
      Wishes.
      
      The genie power to grant wishes is legendary among mortals. Only the
      most potent genies, such as those among the nobility, can do so. A
      particular genie that has this power can grant one to three wishes to
      a creature that isn't a genie. Once a genie has granted its limit of
      wishes, it can't grant wishes again for some amount of time (usually 1
      year). and cosmic law dictates that the same genie can expend its
      limit of wishes on a specific creature only once in that creature's
      existence.
      
      To be granted a wish, a creature within 60 feet of the genie states a
      desired effect to it. The genie can then cast the wish spell on the
      creature's behalf to bring about the effect. Depending on the genie's
      nature, the genie might try to pervert the intent of the wish by
      exploiting the wish's poor wording. The perversion of the wording is
      usually crafted to be to the genie's benefit.
    Multiattack.
      The efreeti makes two scimitar attacks or uses its Hurl Flame twice.
    Scimitar.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one target. Hit: 13 (2d6
      + 6) slashing damage plus 7 (2d6) fire damage.
    Hurl Flame.
      Ranged Spell Attack: +7 to hit, range 120 ft., one target. Hit: 17
      (5d6) fire damage.
    """
    name = "Efreeti"
    description = "Large elemental, lawful evil"
    challenge_rating = 11
    armor_class = 17
    skills = ""
    senses = "Darkvision 120 ft., Passive Perception 12"
    languages = "Ignan"
    strength = Ability(22)
    dexterity = Ability(12)
    constitution = Ability(24)
    intelligence = Ability(16)
    wisdom = Ability(15)
    charisma = Ability(16)
    speed = 40
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 200
    hit_dice = "16d10"
    spells = []


class Elephant(Monster):
    """
    Trampling Charge.
      If the elephant moves at least 20 ft. straight toward a creature and
      then hits it with a gore attack on the same turn, that target must
      succeed on a DC 12 Strength saving throw or be knocked prone. If the
      target is prone, the elephant can make one stomp attack against it as
      a bonus action.
    Gore.
      Melee Weapon Attack: +8 to hit, reach 5 ft., one target. Hit: 19 (3d8
      + 6) piercing damage.
    Stomp.
      Melee Weapon Attack: +8 to hit, reach 5 ft., one prone creature. Hit:
      22 (3d10 + 6) bludgeoning damage.
    """
    name = "Elephant"
    description = "Huge beast, unaligned"
    challenge_rating = 4
    armor_class = 12
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(22)
    dexterity = Ability(9)
    constitution = Ability(17)
    intelligence = Ability(3)
    wisdom = Ability(11)
    charisma = Ability(6)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 76
    hit_dice = "8d12"
    spells = []


class Elk(Monster):
    """
    Charge.
      If the elk moves at least 20 ft. straight toward a target and then
      hits it with a ram attack on the same turn, the target takes an extra
      7 (2d6) damage. If the target is a creature, it must succeed on a DC
      13 Strength saving throw or be knocked prone.
    Ram.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 6 (1d6 +
      3) bludgeoning damage.
    Hooves.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one prone creature. Hit:
      8 (2d4 + 3) bludgeoning damage.
    """
    name = "Elk"
    description = "Large beast, unaligned"
    challenge_rating = 0.25
    armor_class = 10
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(16)
    dexterity = Ability(10)
    constitution = Ability(12)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(6)
    speed = 50
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 13
    hit_dice = "2d10"
    spells = []


class Erinyes(Monster):
    """
    Hellish Weapons.
      The erinyes's weapon attacks are magical and deal an extra 13 (3d8)
      poison damage on a hit (included in the attacks).
    Magic Resistance.
      The erinyes has advantage on saving throws against spells and other
      magical effects.
    Multiattack.
      The erinyes makes three attacks
    Longsword.
      Melee Weapon Attack: +8 to hit, reach 5 ft., one target. Hit: 8 (1d8 +
      4) slashing damage, or 9 (1d10 + 4) slashing damage if used with two
      hands, plus 13 (3d8) poison damage.
    Longbow.
      Ranged Weapon Attack: +7 to hit, range 150/600 ft., one target. Hit: 7
      (1d8 + 3) piercing damage plus 13 (3d8) poison damage, and the target
      must succeed on a DC 14 Constitution saving throw or be poisoned. The
      poison lasts until it is removed by the lesser restoration spell or
      similar magic.
    Variant: Rope of Entanglement.
      Some erinyes carry a rope of entanglement (detailed in the Dungeon
      Master's Guide). When such an erinyes uses its Multiattack, the
      erinyes can use the rope in place of two of the attacks.
    """
    name = "Erinyes"
    description = "Medium fiend, lawful evil"
    challenge_rating = 12
    armor_class = 18
    skills = ""
    senses = "Truesight 120 ft., Passive Perception 12"
    languages = "Infernal, telepathy 120 ft."
    strength = Ability(18)
    dexterity = Ability(16)
    constitution = Ability(18)
    intelligence = Ability(14)
    wisdom = Ability(14)
    charisma = Ability(18)
    speed = 30
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 153
    hit_dice = "18d8"
    spells = []


class Ettercap(Monster):
    """
    Spider Climb.
      The ettercap can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Web Sense.
      While in contact with a web, the ettercap knows the exact location of
      any other creature in contact with the same web.
    Web Walker.
      The ettercap ignores movement restrictions caused by webbing.
    Multiattack.
      The ettercap makes two attacks: one with its bite and one with its
      claws.
    Bite.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one creature. Hit: 6 (1d8
      + 2) piercing damage plus 4 (1d8) poison damage. The target must
      succeed on a DC 11 Constitution saving throw or be poisoned for 1
      minute. The creature can repeat the saving throw at the end of each of
      its turns, ending the effect on itself on a success.
    Claws.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 7 (2d4 +
      2) slashing damage.
    Web.
      Ranged Weapon Attack: +4 to hit, range 30/60 ft., one Large or smaller
      creature. Hit: The creature is restrained by webbing. As an action,
      the restrained creature can make a DC 11 Strength check, escaping from
      the webbing on a success. The effect ends if the webbing is destroyed.
      The webbing has AC 10, 5 hit points, is vulnerable to fire damage and
      immune to bludgeoning damage.
    Variant: Web Garrote.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one Medium or Small
      creature against which the ettercap has advantage on the attack roll.
      Hit: 4 (1d4 + 2) bludgeoning damage, and the target is grappled
      (escape DC 12). Until this grapple ends, the target can't breathe, and
      the ettercap has advantage on attack rolls against it.
    """
    name = "Ettercap"
    description = "Medium monstrosity, neutral evil"
    challenge_rating = 2
    armor_class = 13
    skills = "Perception +3, Stealth +4, Survival +3"
    senses = "Darkvision 60 ft., Passive Perception 13"
    languages = ""
    strength = Ability(14)
    dexterity = Ability(15)
    constitution = Ability(13)
    intelligence = Ability(7)
    wisdom = Ability(12)
    charisma = Ability(8)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 30
    hp_max = 44
    hit_dice = "8d8"
    spells = []


class Ettin(Monster):
    """
    Two Heads.
      The ettin has advantage on Wisdom (Perception) checks and on saving
      throws against being blinded, charmed, deafened, frightened, stunned,
      and knocked unconscious.
    Wakeful.
      When one of the ettin's heads is asleep, its other head is awake.
    Multiattack.
      The ettin makes two attacks: one with its battleaxe and one with its
      morningstar.
    Battleaxe.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 14 (2d8
      + 5) slashing damage.
    Morningstar.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 14 (2d8
      + 5) piercing damage.
    """
    name = "Ettin"
    description = "Large giant, chaotic evil"
    challenge_rating = 4
    armor_class = 12
    skills = "Perception +4"
    senses = "Darkvision 60 ft., Passive Perception 14"
    languages = "Giant, Orc"
    strength = Ability(21)
    dexterity = Ability(8)
    constitution = Ability(17)
    intelligence = Ability(6)
    wisdom = Ability(10)
    charisma = Ability(8)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 85
    hit_dice = "10d10"
    spells = []



from dungeonsheets import weapons, features


def all_features():
    """Generate all the valid features classes so far defined."""
    for feature_name, feature in features.__dict__.items():
        # Check if it's (a) a class and (b) a subclass of ``Features``
        if isinstance(feature, type) and issubclass(feature, Feature):
            yield feature


def create_feature(**params):
    """Create a new subclass of ``Feature`` with given default parameters.

    Useful for features that haven't been entered into the ``features.py``
    file yet.

    Parameters
    ----------
    params : optional
     Saved as attributes of the new class.

    Returns
    -------
    NewFeature
     New feature class, subclass of ``Feature``, with given params.
    """
    NewFeature = type("UnknownFeature", (Feature,), params)
    return NewFeature


class Feature:
    """
    Provide full text of rules in documentation
    """

    name = "Generic Feature"
    owner = None
    source = "Unknown"  # race, class, background, etc.
    spells_known = ()
    spells_prepared = ()
    needs_implementation = False  # Set to True if need to find way to compute stats

    def __init__(self, owner=None):
        self.owner = owner
        self.spells_known = [S() for S in self.spells_known]
        self.spells_prepared = [S() for S in self.spells_prepared]

    def __eq__(self, other):
        return (self.name == other.name) and (self.source == other.source)

    def __hash__(self):
        return 0

    def __str__(self):
        return self.name

    def __repr__(self):
        return '"{:s}"'.format(self.name)

    def weapon_func(self, weapon: weapons.Weapon, **kwargs):
        """
        Updates weapon based on the Feature property

        Parameters
        ----------
        weapon
          The weapon to be tested for special bonuses
        kwargs
          Any other key-word arguments the function may require
        """
        pass


class FeatureSelector(Feature):
    """
    A feature with multiple possible choices.
    """

    options = dict()
    name = ""
    source = ""

    def __new__(t, owner, feature_choices=[]):
        # Look for matching feature_choices
        new_feat = Feature.__new__(Feature, owner=owner)
        new_feat.__doc__ = t.__doc__
        new_feat.name = t.name
        new_feat.source = t.source
        new_feat.needs_implementation = True
        for selection in feature_choices:
            if selection.lower() in t.options:
                feat_class = t.options[selection.lower()]
                if owner.has_feature(feat_class):
                    continue
                new_feat = feat_class(owner=owner)
                new_feat.source = t.source
        return new_feat



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Zombie(Monster):
    """
    Undead Fortitude.
      If damage reduces the zombie to 0 hit points, it must make a
      Constitution saving throw with a DC of 5+the damage taken, unless the
      damage is radiant or from a critical hit. On a success, the zombie
      drops to 1 hit point instead.
    Slam.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 4 (1d6 +
      1) bludgeoning damage.
    """
    name = "Zombie"
    description = "Medium undead, neutral evil"
    challenge_rating = 0.25
    armor_class = 8
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 8"
    languages = "understands all languages it spoke in life but can't speak"
    strength = Ability(13)
    dexterity = Ability(6)
    constitution = Ability(16)
    intelligence = Ability(3)
    wisdom = Ability(6)
    charisma = Ability(5)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 22
    hit_dice = "3d8"
    spells = []



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Tarrasque(Monster):
    """
    Legendary Resistance.
      If the tarrasque fails a saving throw, it can choose to succeed
      instead.
    Magic Resistance.
      The tarrasque has advantage on saving throws against spells and other
      magical effects.
    Reflective Carapace.
      Any time the tarrasque is targeted by a magic missile spell, a line
      spell, or a spell that requires a ranged attack roll, roll a d6. On a
      1 to 5, the tarrasque is unaffected. On a 6, the tarrasque is
      unaffected, and the effect is reflected back at the caster as though
      it originated from the tarrasque, turning the caster into the target.
    Siege Monster.
      The tarrasque deals double damage to objects and structures.
    Multiattack.
      The tarrasque can use its Frightful Presence. It then makes five
      attacks: one with its bite, two with its claws, one with its horns,
      and one with its tail. It can use its Swallow instead of its bite.
    Bite.
      Melee Weapon Attack: +19 to hit, reach 10 ft., one target. Hit: 36
      (4d12 + 10) piercing damage. If the target is a creature, it is
      grappled (escape DC 20). Until this grapple ends, the target is
      restrained, and the tarrasque can't bite another target.
    Claw.
      Melee Weapon Attack: +19 to hit, reach 15 ft., one target. Hit: 28
      (4d8 + 10) slashing damage.
    Horns.
      Melee Weapon Attack: +19 to hit, reach 10 ft., one target. Hit: 32
      (4d10 + 10) piercing damage.
    Tail.
      Melee Weapon Attack: +19 to hit, reach 20 ft., one target. Hit: 24
      (4d6 + 10) bludgeoning damage. If the target is a creature, it must
      succeed on a DC 20 Strength saving throw or be knocked prone.
    Frightful Presence.
      Each creature of the tarrasque's choice within 120 feet of it and
      aware of it must succeed on a DC 17 Wisdom saving throw or become
      frightened for 1 minute. A creature can repeat the saving throw at the
      end of each of its turns, with disadvantage if the tarrasque is within
      line of sight, ending the effect on itself on a success. If a
      creature's saving throw is successful or the effect ends for it, the
      creature is immune to the tarrasque's Frightful Presence for the next
      24 hours.
    Swallow.
      The tarrasque makes one bite attack against a Large or smaller
      creature it is grappling. If the attack hits, the target takes the
      bite's damage, the target is swallowed, and the grapple ends. While
      swallowed, the creature is blinded and restrained, it has total cover
      against attacks and other effects outside the tarrasque, and it takes
      56 (16d6) acid damage at the start of each of the tarrasque's turns.
      
      If the tarrasque takes 60 damage or more on a single turn from a
      creature inside it, the tarrasque must succeed on a DC 20 Constitution
      saving throw at the end of that turn or regurgitate all swallowed
      creatures, which fall prone in a space within 10 feet of the
      tarrasque. If the tarrasque dies, a swallowed creature is no longer
      restrained by it and can escape from the corpse by using 30 feet of
      movement, exiting prone.
    """
    name = "Tarrasque"
    description = "Gargantuan monstrosity, unaligned"
    challenge_rating = 30
    armor_class = 25
    skills = ""
    senses = "Blindsight 120 ft., Passive Perception 10"
    languages = ""
    strength = Ability(30)
    dexterity = Ability(11)
    constitution = Ability(30)
    intelligence = Ability(3)
    wisdom = Ability(11)
    charisma = Ability(11)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 676
    hit_dice = "33d20"
    spells = []


class Thug(Monster):
    """
    Pack Tactics.
      The thug has advantage on an attack roll against a creature if at
      least one of the thug's allies is within 5 ft. of the creature and the
      ally isn't incapacitated.
    Multiattack.
      The thug makes two melee attacks.
    Mace.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one creature. Hit: 5 (1d6
      + 2) bludgeoning damage.
    Heavy Crossbow.
      Ranged Weapon Attack: +2 to hit, range 100/400 ft., one target. Hit: 5
      (1d10) piercing damage.
    """
    name = "Thug"
    description = "Medium humanoid, any non-good alignment"
    challenge_rating = 0.5
    armor_class = 11
    skills = "Intimidation +2"
    senses = "Passive Perception 10"
    languages = "any one language (usually Common)"
    strength = Ability(15)
    dexterity = Ability(11)
    constitution = Ability(14)
    intelligence = Ability(10)
    wisdom = Ability(10)
    charisma = Ability(11)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 32
    hit_dice = "5d8"
    spells = []


class Tiger(Monster):
    """
    Keen Smell.
      The tiger has advantage on Wisdom (Perception) checks that rely on
      smell.
    Pounce.
      If the tiger moves at least 20 ft. straight toward a creature and then
      hits it with a claw attack on the same turn, that target must succeed
      on a DC 13 Strength saving throw or be knocked prone. If the target is
      prone, the tiger can make one bite attack against it as a bonus
      action.
    Bite.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 8 (1d10
      + 3) piercing damage.
    Claw.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 7 (1d8 +
      3) slashing damage.
    """
    name = "Tiger"
    description = "Large beast, unaligned"
    challenge_rating = 1
    armor_class = 12
    skills = "Perception +3, Stealth +6"
    senses = "Darkvision 60 ft., Passive Perception 13"
    languages = ""
    strength = Ability(17)
    dexterity = Ability(15)
    constitution = Ability(14)
    intelligence = Ability(3)
    wisdom = Ability(12)
    charisma = Ability(8)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 37
    hit_dice = "5d10"
    spells = []


class Treant(Monster):
    """
    False Appearance.
      While the treant remains motionless, it is indistinguishable from a
      normal tree.
    Siege Monster.
      The treant deals double damage to objects and structures.
    Multiattack.
      The treant makes two slam attacks.
    Slam.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one target. Hit: 16 (3d6
      + 6) bludgeoning damage.
    Rock.
      Ranged Weapon Attack: +10 to hit, range 60/180 ft., one target. Hit:
      28 (4d10 + 6) bludgeoning damage.
    Animate Trees.
      The treant magically animates one or two trees it can see within 60
      feet of it. These trees have the same statistics as a treant, except
      they have Intelligence and Charisma scores of 1, they can't speak, and
      they have only the Slam action option. An animated tree acts as an
      ally of the treant. The tree remains animate for 1 day or until it
      dies; until the treant dies or is more than 120 feet from the tree; or
      until the treant takes a bonus action to turn it back into an
      inanimate tree. The tree then takes root if possible.
    """
    name = "Treant"
    description = "Huge plant, chaotic good"
    challenge_rating = 9
    armor_class = 16
    skills = ""
    senses = "Passive Perception 13"
    languages = "Common, Druidic, Elvish, Sylvan"
    strength = Ability(23)
    dexterity = Ability(8)
    constitution = Ability(21)
    intelligence = Ability(12)
    wisdom = Ability(16)
    charisma = Ability(12)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 138
    hit_dice = "12d12"
    spells = []


class TribalWarrior(Monster):
    """
    Pack Tactics.
      The warrior has advantage on an attack roll against a creature if at
      least one of the warrior's allies is within 5 ft. of the creature and
      the ally isn't incapacitated.
    Spear.
      Melee or Ranged Weapon Attack: +3 to hit, reach 5 ft. or range 20/60
      ft., one target. Hit: 4 (1d6 + 1) piercing damage, or 5 (1d8 + 1)
      piercing damage if used with two hands to make a melee attack.
    """
    name = "Tribal Warrior"
    description = "Medium humanoid, any alignment"
    challenge_rating = 0.125
    armor_class = 12
    skills = ""
    senses = "Passive Perception 10"
    languages = "any one language"
    strength = Ability(13)
    dexterity = Ability(11)
    constitution = Ability(12)
    intelligence = Ability(8)
    wisdom = Ability(11)
    charisma = Ability(8)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 11
    hit_dice = "2d8"
    spells = []


class Triceratops(Monster):
    """
    Trampling Charge.
      If the triceratops moves at least 20 ft. straight toward a creature
      and then hits it with a gore attack on the same turn, that target must
      succeed on a DC 13 Strength saving throw or be knocked prone. If the
      target is prone, the triceratops can make one stomp attack against it
      as a bonus action.
    Gore.
      Melee Weapon Attack: +9 to hit, reach 5 ft., one target. Hit: 24 (4d8
      + 6) piercing damage.
    Stomp.
      Melee Weapon Attack: +9 to hit, reach 5 ft., one prone creature. Hit:
      22 (3d10 + 6) bludgeoning damage
    """
    name = "Triceratops"
    description = "Huge beast, unaligned"
    challenge_rating = 5
    armor_class = 13
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(22)
    dexterity = Ability(9)
    constitution = Ability(17)
    intelligence = Ability(2)
    wisdom = Ability(11)
    charisma = Ability(5)
    speed = 50
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 95
    hit_dice = "10d12"
    spells = []


class Troll(Monster):
    """
    Keen Smell.
      The troll has advantage on Wisdom (Perception) checks that rely on
      smell.
    Regeneration.
      The troll regains 10 hit points at the start of its turn. If the troll
      takes acid or fire damage, this trait doesn't function at the start of
      the troll's next turn. The troll dies only if it starts its turn with
      0 hit points and doesn't regenerate.
    Variant: Loathsome Limbs.
      Whenever the troll takes at least 15 slashing damage at one time, roll
      a d20 to determine what else happens to it:
      
      1-10: Nothing else happens.
      
      11-14: One leg is severed from the troll if it has any legs left.
      
      15- 18: One arm is severed from the troll if it has any arms left.
      
      19-20: The troll is decapitated, but the troll dies only if it can't
      regenerate. If it dies, so does the severed head.
      
      If the troll finishes a short or long rest without reattaching a
      severed limb or head, the part regrows. At that point, the severed
      part dies. Until then, a severed part acts on the troll's initiative
      and has its own action and movement. A severed part has AC 13, 10 hit
      points, and the troll's Regeneration trait.
      
      A severed leg is unable to attack and has a speed of 5 feet.
      
      A severed arm has a speed of 5 feet and can make one claw attack on
      its turn, with disadvantage on the attack roll unless the troll can
      see the arm and its target. Each time the troll loses an arm, it loses
      a claw attack.
      
      If its head is severed, the troll loses its bite attack and its body
      is blinded unless the head can see it. The severed head has a speed of
      0 feet and the troll's Keen Smell trait. It can make a bite attack but
      only against a target in its space.
      
      The troll's speed is halved if it's missing a leg. If it loses both
      legs, it falls prone. If it has both arms, it can crawl. With only one
      arm, it can still crawl, but its speed is halved. With no arms or
      legs, its speed is 0, and it can't benefit from bonuses to speed.
    Multiattack.
      The troll makes three attacks: one with its bite and two with its
      claws.
    Bite.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 7 (1d6 +
      4) piercing damage.
    Claw.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 11 (2d6
      + 4) slashing damage.
    """
    name = "Troll"
    description = "Large giant, chaotic evil"
    challenge_rating = 5
    armor_class = 15
    skills = "Perception +2"
    senses = "Darkvision 60 ft., Passive Perception 12"
    languages = "Giant"
    strength = Ability(18)
    dexterity = Ability(13)
    constitution = Ability(20)
    intelligence = Ability(7)
    wisdom = Ability(9)
    charisma = Ability(7)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 84
    hit_dice = "8d10"
    spells = []


class TyrannosaurusRex(Monster):
    """
    Multiattack.
      The tyrannosaurus makes two attacks: one with its bite and one with
      its tail. It can't make both attacks against the same target.
    Bite.
      Melee Weapon Attack: +10 to hit, reach 10 ft., one target. Hit: 33
      (4d12 + 7) piercing damage. If the target is a Medium or smaller
      creature, it is grappled (escape DC 17). Until this grapple ends, the
      target is restrained, and the tyrannosaurus can't bite another target.
    Tail.
      Melee Weapon Attack: +10 to hit, reach 10 ft., one target. Hit: 20
      (3d8 + 7) bludgeoning damage.
    """
    name = "Tyrannosaurus Rex"
    description = "Huge beast, unaligned"
    challenge_rating = 8
    armor_class = 13
    skills = "Perception +4"
    senses = "Passive Perception 14"
    languages = ""
    strength = Ability(25)
    dexterity = Ability(10)
    constitution = Ability(19)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(9)
    speed = 50
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 136
    hit_dice = "13d12"
    spells = []



from dungeonsheets.spells.spells import Spell


class Jump(Spell):
    """You touch a creature. The creature's jump distance is tripled until the spell
    ends.
    """

    name = "Jump"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A grasshopper's hind leg"
    duration = "1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Ranger", "Sorcerer", "Wizard")



from dungeonsheets.content_registry import default_content_registry

from dungeonsheets.monsters.monsters import *
from dungeonsheets.monsters.monsters_a import *
from dungeonsheets.monsters.monsters_b import *
from dungeonsheets.monsters.monsters_c import *
from dungeonsheets.monsters.monsters_d import *
from dungeonsheets.monsters.monsters_e import *
from dungeonsheets.monsters.monsters_f import *
from dungeonsheets.monsters.monsters_g import *
from dungeonsheets.monsters.monsters_h import *
from dungeonsheets.monsters.monsters_i import *
from dungeonsheets.monsters.monsters_j import *
from dungeonsheets.monsters.monsters_k import *
from dungeonsheets.monsters.monsters_l import *
from dungeonsheets.monsters.monsters_m import *
from dungeonsheets.monsters.monsters_n import *
from dungeonsheets.monsters.monsters_o import *
from dungeonsheets.monsters.monsters_p import *
from dungeonsheets.monsters.monsters_q import *
from dungeonsheets.monsters.monsters_r import *
from dungeonsheets.monsters.monsters_s import *
from dungeonsheets.monsters.monsters_t import *
from dungeonsheets.monsters.monsters_u import *
from dungeonsheets.monsters.monsters_v import *
from dungeonsheets.monsters.monsters_w import *
from dungeonsheets.monsters.monsters_x import *
from dungeonsheets.monsters.monsters_y import *
from dungeonsheets.monsters.monsters_z import *


default_content_registry.add_module(__name__)



import math
import warnings
from collections import defaultdict

from dungeonsheets import exceptions, features, monsters, weapons
from dungeonsheets.classes.classes import CharClass, SubClass
from dungeonsheets.content_registry import find_content


# PHB
class LandCircle(SubClass):
    """The Circle of the Land is made up of mystics and sages who safeguard
    ancient knowledge and rites through a vast oral tradition. These druids
    meet within sacred circles of trees or standing stones to whisper primal
    secrets in Druidic. The circle's wisest members preside as the chief
    priests of communities that hold to the Old Faith and serve as advisors to
    the rulers of those folk. As a member of this circle, your magic is
    influenced by the land where you were initiated into the circle's
    mysterious rites

    """

    name = "Circle of the Land"
    circle = "land"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.BonusCantrip, features.NaturalRecovery]
    features_by_level[3] = [features.CircleSpells]
    features_by_level[6] = [features.LandsStride]
    features_by_level[10] = [features.NaturesWard]
    features_by_level[14] = [features.NaturesSanctuary]


class MoonCircle(SubClass):
    """Druids of the Circle of the Moon are fierce guardians of the wilds. Their
    order gathers under the full moon to share news and trade warnings. They
    haunt the deepest parts of the wilderness, where they might go for weeks on
    end before crossing paths with another humanoid creature, let alone another
    druid.

    Changeable as the moon, a druid of this circle might prowl as a great cat
    one night, soar over the treetops as an eagle the next day, and crash
    through the undergrowth in bear form to drive off a trespassing
    monster. The wild is in the druid's blood.

    """

    name = "Circle of the Moon"
    circle = "moon"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.CombatWildShape, features.CircleForms]
    features_by_level[6] = [features.PrimalStrike]
    features_by_level[10] = [features.ElementalWildShape]
    features_by_level[14] = [features.ThousandForms]


# XGTE
class DreamsCircle(SubClass):
    """Druids who are members of the Circle of Dreams hail from regions that have
    strong ties to the Feywild and its dreamlike realms. The druids'
    guardianship of the natural world makes for a natural alliance between them
    and good-aligned fey. These druids seek to fill the world with dreamy
    wonder. Their magic mends wounds and brings joy to downcast hearts, and the
    realms they protect are gleaming, fruitful places, where dream and reality
    blur together and where the weary can find rest.

    """

    name = "Circle of Dreams"
    circle = "dreams"
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.BalmOfTheSummerCourt]
    features_by_level[6] = [features.HearthOfMoonlightAndShadow]
    features_by_level[10] = [features.HiddenPaths]
    features_by_level[14] = [features.WalkerInDreams]


class ShepherdCircle(SubClass):
    """Druids of the Circle of the Shepherd commune with the spirits of nature,
    especially the spirits of beasts and the fey, and call to those spirits for
    aid. These druids recognize that all living things play a role in the
    natural world, yet they focus on protecting animals and fey creatures that
    have difficulty defending themselves. Shepherds, as they are known, see
    such creatures as their charges. They ward off monsters that threaten them,
    rebuke hunters who kill more prey than necessary, and prevent civilization
    from encroaching on rare animal habitats and on sites sacred to the
    fey. Many of these druids are happiest far from cities and towns, content
    to spend their days in the company of animals and the fey creatures of the
    wilds.

    Members of this circle become adventurers to oppose forces that threaten
    their charges or to seek knowledge and power that will help them safeguard
    their charges better. Wherever these druids go, the spirits of
    the wilderness are with them

    """

    name = "Circle of the Shepherd"
    circle = "shepherd"
    languages = ("Sylvan",)
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.SpeechOfTheWoods, features.SpiritTotem]
    features_by_level[6] = [features.MightySummoner]
    features_by_level[10] = [features.GuardianSpirit]
    features_by_level[14] = [features.FaithfulSummons]


# GGTR
class SporesCircle(SubClass):
    """Druids of the Circle of Spores find beauty in decay. They see within
    mold and other fungi the ability to transform lifeless material into
    abundant, ableit somewhat strange, life.

    These druids believe that life and death are parts of a grand cycle, with
    one leading to the other and then back again. Death isn't the end of life,
    but instead a change of state that sees life shift into a new form.

    Druids of this circle have a complex relationship with the undead. Unlike
    most other druids, they see nothing inherently wrong with undeath, which
    they consider to be a companion to life and death. But these druids believe
    that the natural cycle is heathiest when each segment of it is vibrant and
    changing. Undead that seek to replace all life with undeath, or that try to
    avoid passing to a final rest, violate the cycle and must be thwarted.
    """

    name = "Circle of Spores"
    circle = "spores"
    features_by_level = defaultdict(list)
    features_by_level[2] = [
        features.CircleSpells,
        features.HaloOfSpores,
        features.SymbioticEntity,
    ]
    features_by_level[6] = [features.FungalInfestation]
    features_by_level[10] = [features.SpreadingSpores]
    features_by_level[14] = [features.FungalBody]


class Druid(CharClass):
    name = "Druid"
    _wild_shapes = ()
    _circle = ""
    hit_dice_faces = 8
    subclass_select_level = 2
    saving_throw_proficiencies = ("intelligence", "wisdom")
    primary_abilities = ("wisdom",)
    languages = "Druidic"
    _proficiencies_text = (
        "Light armor",
        "medium armor",
        "shields (druids will not wear armor or use shields made of metal)",
        "clubs",
        "daggers",
        "darts",
        "javelins",
        "maces",
        "quarterstaffs",
        "scimitars",
        "sickles",
        "slings",
        "spears",
    )
    weapon_proficiencies = (
        weapons.Club,
        weapons.Dagger,
        weapons.Dart,
        weapons.Javelin,
        weapons.Mace,
        weapons.Quarterstaff,
        weapons.Scimitar,
        weapons.Sickle,
        weapons.Sling,
        weapons.Spear,
    )
    multiclass_weapon_proficiencies = ()
    _multiclass_proficiencies_text = (
        "Light armor",
        "medium armor",
        "shields (druids will not wear armor or use shields made of metal)",
    )
    class_skill_choices = (
        "Arcana",
        "Animal Handling",
        "Insight",
        "Medicine",
        "Nature",
        "Perception",
        "Religion",
        "Survival",
    )
    features_by_level = defaultdict(list)
    features_by_level[2] = [features.WildShape]
    features_by_level[18] = [features.TimelessBody, features.BeastSpells]
    features_by_level[20] = [features.Archdruid]
    subclasses_available = (
        LandCircle,
        MoonCircle,
        DreamsCircle,
        ShepherdCircle,
        SporesCircle,
    )
    spellcasting_ability = "wisdom"
    spell_slots_by_level = {
        1: (2, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        2: (2, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        3: (2, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        4: (3, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        5: (3, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        6: (3, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        7: (3, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        8: (3, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        9: (3, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        10: (4, 4, 3, 3, 3, 2, 0, 0, 0, 0),
        11: (4, 4, 3, 3, 3, 2, 1, 0, 0, 0),
        12: (4, 4, 3, 3, 3, 2, 1, 0, 0, 0),
        13: (4, 4, 3, 3, 3, 2, 1, 1, 0, 0),
        14: (4, 4, 3, 3, 3, 2, 1, 1, 0, 0),
        15: (4, 4, 3, 3, 3, 2, 1, 1, 1, 0),
        16: (4, 4, 3, 3, 3, 2, 1, 1, 1, 0),
        17: (4, 4, 3, 3, 3, 2, 1, 1, 1, 1),
        18: (4, 4, 3, 3, 3, 3, 1, 1, 1, 1),
        19: (4, 4, 3, 3, 3, 3, 2, 1, 1, 1),
        20: (4, 4, 3, 3, 3, 3, 2, 2, 1, 1),
    }

    def select_subclass(self, subclass_str):
        if subclass_str in ["", "None", "none", None]:
            return None
        for sc in self.subclasses_available:
            if (subclass_str.lower() == sc.circle.lower()) or (
                subclass_str.lower() in sc.name.lower()
            ):
                return sc(owner=self.owner)
        return None

    @property
    def circle(self):
        if isinstance(self.subclass, SubClass):
            return self.subclass.circle.lower()
        else:
            return self._circle

    @circle.setter
    def circle(self, circle_str):
        if isinstance(self.subclass, SubClass):
            self.subclass = self.select_subclass(circle_str)
        else:
            self._circle = circle_str

    @property
    def all_wild_shapes(self):
        """Return all wild shapes, regardless of validity."""
        return self._wild_shapes

    @property
    def wild_shapes(self):
        """Return a list of valid wild shapes for this Druid."""
        valid_shapes = []
        for shape in self._wild_shapes:
            # Check if shape can be transformed into
            if self.can_assume_shape(shape):
                valid_shapes.append(shape)
        return valid_shapes

    @wild_shapes.setter
    def wild_shapes(self, new_shapes):
        actual_shapes = []
        # Retrieve the actual monster classes if possible
        for shape in new_shapes:
            if isinstance(shape, monsters.Monster):
                # Already a monster shape so just add it as is
                new_shape = shape
            else:
                # Not already a monster so see if we can find one
                try:
                    NewMonster = find_content(shape, valid_classes=[monsters.Monster])
                    new_shape = NewMonster()
                except exceptions.ContentNotFound:
                    msg = (
                        f"Wild shape '{shape}' not found. Please add it to"
                        " ``monsters.py``"
                    )
                    raise exceptions.MonsterError(msg)
            actual_shapes.append(new_shape)
        # Save the updated list for later
        self._wild_shapes = actual_shapes

    def can_assume_shape(self, shape: monsters.Monster) -> bool:
        """Determine if a given shape meets the requirements for transforming.

        See Pg 66 of player's handbook.

        Parameters
        ==========
        shape
          A monster that the Druid wishes to transform into.

        Returns
        =======
        can_assume
          True if the monster meets the C/R, swim and flying speed
          restrictions.

        """
        # Determine acceptable states based on druid level
        if self.level < 2:
            max_cr = -1
            max_swim = 0
            max_fly = 0
        elif self.level < 4:
            max_cr = 1 / 4
            max_swim = 0
            max_fly = 0
        elif self.level < 8:
            max_cr = 1 / 2
            max_swim = None
            max_fly = 0
        else:
            max_cr = 1
            max_swim = None
            max_fly = None
        # Make adjustments for moon circle druids
        if self.circle == "moon":
            if 2 <= self.level < 6:
                max_cr = 1
            elif self.level >= 6:
                max_cr = math.floor(self.level / 3)
        # Check if the beast shape can be assumed
        valid_cr = max_cr is None or shape.challenge_rating <= max_cr
        valid_swim = max_swim is None or shape.swim_speed <= max_swim
        valid_fly = max_fly is None or shape.fly_speed <= max_fly
        can_assume = shape.is_beast and valid_cr and valid_swim and valid_fly
        if shape.name == "Spider":
            assert can_assume
        return can_assume

    @property
    def spells(self):
        return tuple(S() for S in self.spells_prepared)

    @spells.setter
    def spells(self, val):
        if len(val) > 0:
            warnings.warn(
                "Druids cannot learn spells, use ``spells_prepared`` instead.",
                RuntimeWarning,
            )



#!/usr/bin/env python

"""Launch a system to interactively create a character."""

import logging
import math
import os
from random import randint
import subprocess

import npyscreen

from dungeonsheets import character, race, dice, background, classes, weapons, armor

log = logging.getLogger(__name__)


def read_version():
    version = open(os.path.join(os.path.dirname(__file__), "../VERSION")).read()
    version = version.replace("\n", "")
    return version


char_classes = {c.name: c for c in classes.available_classes}

races = {r.name: r for r in race.available_races}

backgrounds = {b.name: b for b in background.available_backgrounds}

all_weapons = (
    weapons.simple_melee_weapons
    + weapons.martial_melee_weapons
    + weapons.simple_ranged_weapons
    + weapons.martial_ranged_weapons
)


class LinkedListForm(npyscreen.ActionForm):
    prev_page = None
    this_page = None
    next_page = None

    def __init__(self, formid, *args, **kwargs):
        self.this_page = formid
        super().__init__(*args, **kwargs)

    def to_next(self):
        self.parentApp.setNextForm(self.next_page)

    def to_prev(self):
        self.parentApp.setNextForm(self.prev_page)

    def add_next(self, next_name):
        new_next = self.parentApp.getForm(next_name)
        if self.next_page:
            current_next = self.parentApp.getForm(self.next_page)
            current_next.prev_page = next_name
        new_next.next_page = self.next_page
        new_next.prev_page = self.this_page
        self.next_page = next_name

    def add_prev(self, prev_name):
        new_prev = self.parentApp.getForm(prev_name)
        if self.prev_page:
            current_prev = self.parentApp.getForm(self.prev_page)
            current_prev.next_page = prev_name
        new_prev.prev_page = self.prev_page
        new_prev.next_page = self.this_page
        self.prev_page = prev_name

    def prune(self):
        if self.next_page:
            next_form = self.parentApp.getForm(self.next_page)
            next_form.prev_page = self.prev_page
        if self.prev_page:
            prev_form = self.parentApp.getForm(self.prev_page)
            prev_form.next_page = self.next_page
        self.parentApp.removeForm(self.this_page)


class App(npyscreen.NPSAppManaged):
    # STARTING_FORM = 'SKILLS'
    character = None
    n_classes = 1

    def save_character(self):
        # Save the file
        filename = self.getForm("SAVE").filename.value
        self.character.save(filename, template_file="empty_template.txt")
        # Create the PDF character sheet
        if self.getForm("SAVE").make_pdf.value:
            log.debug("Creating PDF")
            self.character.to_pdf(filename)
            subprocess.call(["makesheets", filename])

    def update_max_hp_roll(self):
        abil = self.getForm("ABILITIES")
        # Update max HP based on the class
        hit_dice = [dice.read_dice_str(d) for d in self.character.hit_dice.split(" + ")]
        const = int((int(abil.constitution.value) - 10) / 2)
        hp_max = f"{hit_dice[0].faces + self.character.level*const}"
        hds = {}
        for i, hd in enumerate(hit_dice):
            num = hd.num
            if i == 0:
                num -= 1
            if hd.faces not in hds:
                hds[hd.faces] = 0
            hds[hd.faces] += num
        for faces, num in hds.items():
            hp_max += f" + {num}d{faces}"
        abil.hp_roll_text.value = "Enter (or roll) your Max HP: " + hp_max
        abil.display()

    def reroll_max_hp(self):
        abil = self.getForm("ABILITIES")
        # Update max HP based on the class
        hit_dice = [dice.read_dice_str(d) for d in self.character.hit_dice.split(" + ")]
        const = int((int(abil.constitution.value) - 10) / 2)
        # Assume first hd given is from primary class
        hp_max = hit_dice[0].faces + const
        for i, hd in enumerate(hit_dice):
            num = hd.num
            if i == 0:
                num -= 1
            for d in range(num):
                hp_max += randint(1, hd.faces) + const
        abil.hp_max.value = str(hp_max)
        abil.display()

    def set_default_hp_max(self):
        abil = self.getForm("ABILITIES")
        # Update max HP based on the class
        hit_dice = [dice.read_dice_str(d) for d in self.character.hit_dice.split(" + ")]
        const = int((int(abil.constitution.value) - 10) / 2)
        # Assume first hd given is from primary class
        hp_max = hit_dice[0].faces + const
        for i, hd in enumerate(hit_dice):
            num = hd.num
            if i == 0:
                num -= 1
            for d in range(num):
                hp_max += math.ceil(hd.faces / 2) + const
        log.debug("Updating max hp: %d", hp_max)
        abil.hp_max.value = str(hp_max)
        abil.display()

    def onStart(self):
        self.character = character.Character()
        self.character.class_list = []
        self.addForm("MAIN", BasicInfoForm, name="Basic Info:", formid="MAIN")
        self.addForm(
            "RACE", RaceForm, name="Select your character's race:", formid="RACE"
        )
        self.addForm(
            "CLASS1",
            CharacterClassForm,
            name="Select your character's primary class:",
            formid="CLASS1",
        )
        self.addForm(
            "BACKGROUND", BackgroundForm, name="Choose background:", formid="BACKGROUND"
        )
        self.addForm(
            "ALIGNMENT",
            AlignmentForm,
            name="Select your character's alignment:",
            formid="ALIGNMENT",
        )
        self.addForm(
            "ABILITIES",
            AbilityScoreForm,
            name="Choose ability scores:",
            formid="ABILITIES",
        )
        self.addForm(
            "SKILLS", SkillForm, name="Choose skill proficiencies", formid="SKILLS"
        )
        self.addForm("WEAPONS", WeaponForm, name="Choose weapons", formid="WEAPONS")
        self.addForm("ARMOR", ArmorForm, name="Choose armor", formid="ARMOR")
        self.addForm(
            "PERSONALITY",
            PersonalityForm,
            name="Describe your character",
            formid="PERSONALITY",
        )
        self.addForm("SAVE", SaveForm, name="Save character:", formid="SAVE")

        # Initialized the DoublyLinkedList
        forms = [
            "MAIN",
            "RACE",
            "CLASS1",
            "BACKGROUND",
            "ALIGNMENT",
            "ABILITIES",
            "SKILLS",
            "WEAPONS",
            "ARMOR",
            "PERSONALITY",
            "SAVE",
        ]
        for i in range(len(forms) - 1):
            form = self.getForm(forms[i])
            form.add_next(forms[i + 1])


class BasicInfoForm(LinkedListForm):
    def create(self):
        self.name = self.add(
            npyscreen.TitleText, name="Character Name:", use_two_lines=False
        )
        self.player_name = self.add(
            npyscreen.TitleText, name="Player Name:", use_two_lines=False
        )

    def on_ok(self):
        # Update the default filename
        name = self.name.value or "New Character"
        if name == "":
            filename = "new_character.py"
        else:
            filename = f'{name.split(" ")[0].lower()}.py'
        save_form = self.parentApp.getForm("SAVE")
        save_form.filename.value = filename
        self.parentApp.character.name = self.name.value
        self.parentApp.character.player_name = self.player_name.value
        super().to_next()

    def on_cancel(self):
        raise KeyboardInterrupt


class RaceForm(LinkedListForm):
    def create(self):
        self.race = self.add(
            npyscreen.TitleSelectOne, name="Race:", values=tuple(races.keys())
        )

    def on_ok(self):
        if len(self.race.get_selected_objects()) >= 1:
            selected_race = self.race.get_selected_objects()[0]
            SelectedRace = races[selected_race]
            log.debug("Selected character race: %s", SelectedRace.name)
            self.parentApp.character.race = SelectedRace
            super().to_next()

    def on_cancel(self):
        super().to_prev()


class CharacterClassForm(LinkedListForm):
    class_num = 1

    def __init__(self, num=1, **kwargs):
        self.class_num = num
        self.class_options = list(char_classes.keys())
        super().__init__(**kwargs)

    @property
    def subclass_page(self):
        key = "SUBCLASS{:d}".format(self.class_num)
        if key in self.parentApp._Forms:
            return self.parentApp.getForm(key)
        else:
            return None

    @property
    def next_multiclass_page(self):
        key = "CLASS{:d}".format(self.class_num + 1)
        if key in self.parentApp._Forms:
            return self.parentApp.getForm(key)
        else:
            return None

    def update_options(self):
        if len(self.parentApp.character.class_list) == 0:
            return
        else:
            self.class_options = list(char_classes.keys())
            for c in self.parentApp.character.class_list[: self.class_num - 1]:
                self.class_options.remove(c.name)
            self.character_class.values = sorted(tuple(self.class_options))
            self.character_class.update()

    def create(self):
        if self.class_num > 1:
            self.add(
                npyscreen.FixedText,
                editable=False,
                value="Current Classes: {}".format(self.parentApp.character.name),
            )
        if self.class_num == 1:
            t = "Primary Class:"
        else:
            t = "Class #{:d}:".format(self.class_num)
        for c in self.parentApp.character.class_list:
            self.class_options.remove(c.name)
        if self.class_num == 1:
            self.multiclass = self.add(
                npyscreen.Checkbox,
                name="Add Multiclass?",
                value=False,
            )
        else:
            self.multiclass = self.add(
                npyscreen.Checkbox,
                name="Add Class #{:d}?".format(self.class_num + 1),
                value=False,
            )
        self.level = self.add(
            npyscreen.TitleText, name="Level:", value="1", use_two_lines=False
        )
        self.character_class = self.add(
            npyscreen.TitleSelectOne, name=t, values=tuple(self.class_options)
        )

    def add_multiclass_page(self):
        new_name = "CLASS{:d}".format(self.class_num + 1)
        new_form = self.parentApp.addForm(
            new_name,
            CharacterClassForm,
            name="Select your character's Class #{:d}:".format(self.class_num + 1),
            num=self.class_num + 1,
            formid=new_name,
        )
        self.add_next(new_name)
        return new_form

    def add_subclass_page(self, newclass, level):
        new_name = "SUBCLASS{:d}".format(self.class_num)
        new_form = self.parentApp.addForm(
            new_name,
            SubclassForm,
            name="Select your {:s} Subclass".format(newclass.name),
            newclass=newclass,
            level=level,
            num=self.class_num,
            formid=new_name,
        )
        self.add_next(new_name)
        return new_form

    def on_ok(self):
        if len(self.character_class.get_selected_objects()) >= 1:
            selected_class = self.character_class.get_selected_objects()[0]
            selected_class = char_classes[selected_class]
            log.debug("Selected character class %s", selected_class.name)
            # replace later classes if we've backed up
            self.parentApp.character.class_list = self.parentApp.character.class_list[
                : self.class_num - 1
            ]
            self.parentApp.character.add_class(
                cls=selected_class, level=int(self.level.value), subclass=None
            )
            # add multiclass page if not exists yet
            if self.multiclass.value:
                if self.next_multiclass_page is None:
                    self.add_multiclass_page()
                else:
                    self.next_multiclass_page.update_options()
            else:
                # in case returned a page, prune any future multiclasses
                while self.next_page != "BACKGROUND":
                    f = self.parentApp.getForm(self.next_page)
                    f.prune()
            if int(self.level.value) >= selected_class.subclass_select_level:
                if not self.subclass_page:
                    self.add_subclass_page(
                        newclass=selected_class, level=int(self.level.value)
                    )
            else:
                if self.subclass_page is not None:
                    f = self.parentApp.getForm(self.next_page)
                    f.prune()
            super().to_next()

    def on_cancel(self):
        super().to_prev()


class SubclassForm(LinkedListForm):
    def __init__(self, newclass, level, num=1, **kwargs):
        self.class_num = num
        self.parent_class = newclass
        if len(newclass.subclasses_available) > 0:
            self.subclass_options = [sc.name for sc in newclass.subclasses_available]
        else:
            self.subclass_options = ("None",)
        self.level = level
        super().__init__(**kwargs)

    def create(self):
        self.subclass = self.add(
            npyscreen.TitleSelectOne,
            name="Subclass:",
            values=tuple(self.subclass_options),
        )

    def on_ok(self):
        if len(self.subclass.get_selected_objects()) >= 1:
            sc = self.subclass.get_selected_objects()[0]
            if sc in [None, "", "None"]:
                sc = None
            self.parentApp.character.class_list = self.parentApp.character.class_list[
                : self.class_num - 1
            ]
            self.parentApp.character.add_class(
                cls=self.parent_class, level=self.level, subclass=sc
            )
            super().to_next()

    def on_cancel(self):
        super().to_prev()


class BackgroundForm(LinkedListForm):
    def create(self):
        self.background = self.add(
            npyscreen.TitleSelectOne,
            name="Background:",
            values=tuple(backgrounds.keys()),
        )

    def on_ok(self):
        if len(self.background.get_selected_objects()) >= 1:
            selected_bg = self.background.get_selected_objects()[0]
            Background = backgrounds[selected_bg]
            self.parentApp.character.background = Background()
            # Update the languages based on background and race
            race_languages = self.parentApp.character.race.languages
            languages = Background.languages + race_languages
            self.parentApp.character.languages = ", ".join(languages)
            log.debug("Selected character background: %s", Background.name)
            super().to_next()

    def on_cancel(self):
        super().to_prev()


class AlignmentForm(LinkedListForm):
    """Choose your character's alignment."""

    alignments = (
        "Lawful good",
        "Neutral good",
        "Chaotic good",
        "Lawful neutral",
        "True neutral",
        "Chaotic neutral",
        "Lawful evil",
        "Neutral evil",
        "Chaotic evil",
    )

    def create(self):
        self.alignment = self.add(
            npyscreen.TitleSelectOne, name="Alignment:", values=self.alignments
        )

    def on_ok(self):
        if len(self.alignment.get_selected_objects()) >= 1:
            selected_alignment = self.alignment.get_selected_objects()[
                0
            ]  # values[self.alignment.value]
            log.debug("Selected character alignment %s", selected_alignment)
            self.parentApp.character.alignment = selected_alignment
            # prep additions to abilities page
            self.parentApp.getForm("ABILITIES").prep()
            super().to_next()

    def on_cancel(self):
        super().to_prev()


class AbilityScoreForm(LinkedListForm):
    num_rolls = 0

    def roll_dice(self):
        """Get six ability scores that can then be assigned to abilities."""

        def roll_score():
            # Roll 4 dice and add the 3 highest
            rolls = (randint(1, 6) for i in range(4))
            score = sum(sorted(rolls)[-3:])
            return score

        scores = (roll_score() for i in range(6))
        return tuple(sorted(scores, reverse=True))

    def default_array(self):
        return (15, 14, 13, 12, 10, 8)

    def reroll(self, widget=None):
        self.num_rolls += 1
        new_scores = self.roll_dice()
        self.score_options.value = str(new_scores)[1:-1]
        self.score_options.update()
        self.reroll_button.value = False
        self.reroll_button.name = "Reroll"
        self.reroll_button.update()
        self.default_button.value = False
        self.default_button.update()

    def set_default(self, widget=None):
        new_scores = self.default_array()
        self.score_options.value = str(new_scores)[1:-1]
        self.score_options.update()
        self.default_button.value = True
        self.default_button.update()

    def reroll_hp(self, widget=None):
        self.parentApp.reroll_max_hp()
        self.parentApp.update_max_hp_roll()

    def create(self):
        self.roll_text = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Take the six rolls below and assign each one to an ability.",
        )
        self.score_options = self.add(
            npyscreen.TitleFixedText,
            name="Rolls:",
            editable=False,
            value=str(self.default_array())[1:-1],
        )
        self.default_button = self.add(
            npyscreen.MiniButtonPress,
            name="Use Default Rolls",
            when_pressed_function=self.set_default,
        )
        self.reroll_button = self.add(
            npyscreen.MiniButtonPress, name="Reroll", when_pressed_function=self.reroll
        )

    def prep(self):
        attrs = (
            "strength",
            "dexterity",
            "constitution",
            "intelligence",
            "wisdom",
            "charisma",
        )
        self.class_text = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Key stats for your primary class {:s} are listed with **".format(
                self.parentApp.character.primary_class.name
            ),
        )
        self.race_text = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Do not add racial bonuses, they will be added for you as listed.",
        )
        for attr in attrs:
            if attr in self.parentApp.character.primary_class.primary_abilities:
                name = "**" + attr
            else:
                name = "" + attr
            race_bonus = getattr(self.parentApp.character.race, f"{attr}_bonus")
            if race_bonus != 0:
                name += "({:+d})".format(race_bonus)
            name += ":"
            new_fld = self.add(
                npyscreen.TitleText, name=name, begin_entry_at=24, value="10"
            )
            setattr(self, attr, new_fld)
        self.hp_roll_text = self.add(npyscreen.FixedText, editable=False, value="")
        self.hp_reroll_buttom = self.add(
            npyscreen.MiniButtonPress,
            name="Reroll Max HP",
            when_pressed_function=self.reroll_hp,
        )
        self.hp_max = self.add(npyscreen.TitleText, name="Max HP:")
        self.parentApp.update_max_hp_roll()
        self.parentApp.set_default_hp_max()

    def on_ok(self):
        attrs = (
            "strength",
            "dexterity",
            "constitution",
            "intelligence",
            "wisdom",
            "charisma",
            "hp_max",
        )
        for attr in attrs:
            fld = getattr(self, attr)
            race_bonus = getattr(self.parentApp.character.race, f"{attr}_bonus", 0)
            try:
                val = int(float(fld.value))
            except ValueError:
                # Not an integer, so clear the field
                val = 10
            val += race_bonus
            log.debug("Setting %s to %s", attr, str(val))
            # Update the "character" with new values
            setattr(self.parentApp.character, attr, val)
        super().to_next()

    def on_cancel(self):
        super().to_prev()


class SkillForm(LinkedListForm):
    def while_editing(self):
        # Update the static skills for race and background
        bg_skills = self.parentApp.character.background.skill_proficiencies
        self.bg_skills.value = str(bg_skills)[1:-1].replace("'", "")
        race_skills = self.parentApp.character.race.skill_proficiencies
        self.race_skills.value = str(race_skills)[1:-1].replace("'", "")
        # Now set the available discretionary choices
        choices = (
            self.parentApp.character.primary_class.class_skill_choices
            + self.parentApp.character.race.skill_choices
            + self.parentApp.character.background.skill_choices
        )
        static_skills = bg_skills + race_skills
        choices = set([c for c in choices if c.lower() not in static_skills])
        self.skill_proficiencies.set_values(sorted(tuple(choices)))
        self.update_remaining()

    def update_remaining(self, widget=None):
        num_choices = (
            self.parentApp.character.primary_class.num_skill_choices
            + self.parentApp.character.race.num_skill_choices
            + self.parentApp.character.background.num_skill_choices
        )
        num_selected = len(self.skill_proficiencies.value)
        remaining = num_choices - num_selected
        log.debug(f"Remaining: {remaining}")
        self.remaining.value = str(remaining)
        self.display()

    def create(self):
        self.bg_skills = self.add(
            npyscreen.TitleText, name="Background:", value="", editable=False
        )
        self.race_skills = self.add(
            npyscreen.TitleText, name="Racial:", value="", editable=False
        )
        self.remaining = self.add(
            npyscreen.TitleText, name="Remaining:", value=0, editable=False
        )
        self.skill_proficiencies = self.add(
            npyscreen.TitleMultiSelect,
            name="Skill Proficiencies:",
            values=(),
            value_changed_callback=self.update_remaining,
        )

    def on_ok(self):
        new_skills = self.skill_proficiencies.get_selected_objects()
        if new_skills is not None:
            new_skills = tuple(s.lower() for s in new_skills)
        else:
            new_skills = ()
        bg_skills = tuple(self.parentApp.character.background.skill_proficiencies)
        race_skills = tuple(self.parentApp.character.race.skill_proficiencies)
        all_skills = new_skills + bg_skills + race_skills
        self.parentApp.character.skill_proficiencies = all_skills
        self.parentApp.getForm("WEAPONS").update_options()
        log.debug(f"Skill proficiencies: {all_skills}")
        super().to_next()

    def on_cancel(self):
        super().to_prev()


class WeaponForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText, editable=False, value="Please select your weapons."
        )
        self.remaining = self.add(
            npyscreen.TitleText, name="Remaining:", value=3, editable=False
        )
        self.weapons = self.add(
            npyscreen.TitleMultiSelect,
            name="Weapons:",
            values=tuple([wpn.name for wpn in all_weapons]),
            value_changed_callback=self.update_remaining,
        )

    def on_ok(self):
        new_weapons = self.weapons.get_selected_objects()
        if new_weapons is not None:
            new_weapons = tuple(s.lower() for s in new_weapons)
        else:
            new_weapons = ()
        for wpn in new_weapons:
            self.parentApp.character.wield_weapon(wpn)
        log.debug(f"Weapons wielded: {new_weapons}")
        super().to_next()

    def update_remaining(self, widget=None):
        num_choices = 3
        num_selected = len(self.weapons.value)
        remaining = num_choices - num_selected
        log.debug(f"Remaining: {remaining}")
        self.remaining.value = str(remaining)
        self.display()

    def update_options(self):
        available_weapons = []
        for wpn in all_weapons:
            if self.parentApp.character.is_proficient(wpn()):
                available_weapons.append(wpn.name)
        self.weapons.values = available_weapons
        self.parentApp.getForm("ARMOR").update_options()
        self.display()

    def on_cancel(self):
        super().to_prev()


class ArmorForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText, editable=False, value="Please select your armor."
        )
        self.shield = self.add(npyscreen.Checkbox, name="Shield? ", value=False)
        self.armor = self.add(
            npyscreen.TitleSelectOne,
            name="Armor:",
            values=tuple([a.name for a in ([armor.NoArmor] + armor.all_armors)]),
        )

    def on_ok(self):
        my_armor = self.armor.get_selected_objects()[0]
        if my_armor.lower() != "no armor":
            self.parentApp.character.wear_armor(my_armor)
        if self.shield.value:
            self.parentApp.character.wield_shield("shield")
        super().to_next()

    def update_options(self):
        available_armors = [armor.NoArmor]
        proficiencies = self.parentApp.character.proficiencies_text.lower()
        if ("light armor" in proficiencies) or ("all armor" in proficiencies):
            available_armors.extend(armor.light_armors)
        if ("medium armor" in proficiencies) or ("all armor" in proficiencies):
            available_armors.extend(armor.medium_armors)
        if ("heavy armor" in proficiencies) or ("all armor" in proficiencies):
            available_armors.extend(armor.heavy_armors)
        self.armor.values = [a.name for a in available_armors]
        self.display()

    def on_cancel(self):
        super().to_prev()


class PersonalityForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Please describe your character's personality.",
        )
        self.add(npyscreen.FixedText, editable=False, value=" ")
        self.pers_instr = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Describe how your character behaves/interacts with others?",
        )
        self.personality_traits = self.add(
            npyscreen.TitleText, name="Personality Traits: ", begin_entry_at=24
        )
        self.add(npyscreen.FixedText, editable=False, value=" ")
        self.ideal_instr = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Describe what values does your character believes in?",
        )
        self.ideals = self.add(npyscreen.TitleText, name="Ideals: ", begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False, value=" ")
        self.bond_instr = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Describe your character's commitments or ongoing quests.",
        )
        self.bonds = self.add(npyscreen.TitleText, name="Bonds: ", begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False, value=" ")
        self.flaw_instr = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Describe your character's interesting flaws.",
        )
        self.flaws = self.add(npyscreen.TitleText, name="Flaws: ", begin_entry_at=24)
        self.add(npyscreen.FixedText, editable=False, value=" ")
        self.feat_instr = self.add(
            npyscreen.FixedText,
            editable=False,
            value="Describe any other notable features or abilities.",
        )
        self.features = self.add(
            npyscreen.TitleText, name="Features: ", begin_entry_at=24
        )

    def on_ok(self):
        if self.personality_traits.value:
            self.parentApp.character.personality_traits = self.personality_traits.value
        if self.ideals.value:
            self.parentApp.character.ideals = self.ideals.value
        if self.bonds.value:
            self.parentApp.character.bonds = self.bonds.value
        if self.flaws.value:
            self.parentApp.character.flaws = self.flaws.value
        if self.features.value:
            self.parentApp.character.features_and_traits = self.features.value
        super().to_next()

    def on_cancel(self):
        super().to_prev()


class SaveForm(LinkedListForm):
    def create(self):
        self.instructions = self.add(
            npyscreen.FixedText,
            editable=False,
            value=(
                "Your character will be saved in the file given below. "
                "After saving, edit this file to finish your personality, "
                "weapons, etc."
            ),
        )
        self.filename = self.add(npyscreen.TitleText, name="Filename:")
        self.make_pdf = self.add(npyscreen.Checkbox, name="Create PDF:", value=True)

    def on_ok(self):
        super().to_next()

    def on_cancel(self):
        super().to_prev()


def main():
    my_app = App()

    try:
        my_app.run()
    except KeyboardInterrupt:
        log.error("Aborted by user request")
    else:
        my_app.save_character()


if __name__ == "__main__":
    main()



from dungeonsheets.features.features import Feature, create_feature, all_features

from dungeonsheets.features.artificer import *
from dungeonsheets.features.backgrounds import *
from dungeonsheets.features.barbarian import *
from dungeonsheets.features.bard import *
from dungeonsheets.features.cleric import *
from dungeonsheets.features.druid import *
from dungeonsheets.features.feats import *
from dungeonsheets.features.fighter import *
from dungeonsheets.features.monk import *
from dungeonsheets.features.paladin import *
from dungeonsheets.features.races import *
from dungeonsheets.features.ranger import *
from dungeonsheets.features.rogue import *
from dungeonsheets.features.sorceror import *
from dungeonsheets.features.warlock import *
from dungeonsheets.features.wizard import *

from dungeonsheets.content_registry import default_content_registry


default_content_registry.add_module(__name__)



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Vampire(Monster):
    """
    Shapechanger.
      If the vampire isn't in sun light or running water, it can use its
      action to polymorph into a Tiny bat or a Medium cloud of mist, or back
      into its true form.
      
      While in bat form, the vampire can't speak, its walking speed is 5
      feet, and it has a flying speed of 30 feet. Its statistics, other than
      its size and speed, are unchanged. Anything it is wearing transforms
      with it, but nothing it is carrying does. It reverts to its true form
      if it dies.
      
      While in mist form, the vampire can't take any actions, speak, or
      manipulate objects. It is weightless, has a flying speed of 20 feet,
      can hover, and can enter a hostile creature's space and stop there. In
      addition, if air can pass through a space, the mist can do so without
      squeezing, and it can't pass through water. It has advantage on
      Strength, Dexterity, and Constitution saving throws, and it is immune
      to all nonmagical damage, except the damage it takes from sunlight.
    Legendary Resistance.
      If the vampire fails a saving throw, it can choose to succeed instead.
    Misty Escape.
      When it drops to 0 hit points outside its resting place, the vampire
      transforms into a cloud of mist (as in the Shapechanger trait) instead
      of falling unconscious, provided that it isn't in sunlight or running
      water. If it can't transform, it is destroyed.
      
      While it has 0 hit points in mist form, it can't revert to its vampire
      form, and it must reach its resting place within 2 hours or be
      destroyed. Once in its resting place, it reverts to its vampire form.
      It is then paralyzed until it regains at least 1 hit point. After
      spending 1 hour in its resting place with 0 hit points, it regains 1
      hit point.
    Regeneration.
      The vampire regains 20 hit points at the start of its turn if it has
      at least 1 hit point and isn't in sunlight or running water. If the
      vampire takes radiant damage or damage from holy water, this trait
      doesn't function at the start of the vampire's next turn.
    Spider Climb.
      The vampire can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Vampire Weaknesses.
      The vampire has the following flaws:
      
      Forbiddance. The vampire can't enter a residence without an invitation
      from one of the occupants.
      
      Harmed by Running Water. The vampire takes 20 acid damage if it ends
      its turn in running water.
      
      Stake to the Heart. If a piercing weapon made of wood is driven into
      the vampire's heart while the vampire is incapacitated in its resting
      place, the vampire is paralyzed until the stake is removed.
      
      Sunlight Hypersensitivity. The vampire takes 20 radiant damage when it
      starts its turn in sunlight. While in sunlight, it has disadvantage on
      attack rolls and ability checks.
    Multiattack (Vampire Form Only).
      The vampire makes two attacks, only one of which can be a bite attack.
    Unarmed Strike (Vampire Form Only).
      Melee Weapon Attack: +9 to hit, reach 5 ft., one creature. Hit: 8 (1d8
      + 4) bludgeoning damage. Instead of dealing damage, the vampire can
      grapple the target (escape DC 18).
    Bite (Bat or Vampire Form Only).
      Melee Weapon Attack: +9 to hit, reach 5 ft., one willing creature, or
      a creature that is grappled by the vampire, incapacitated, or
      restrained. Hit: 7 (1d6 + 4) piercing damage plus 10 (3d6) necrotic
      damage. The target's hit point maximum is reduced by an amount equal
      to the necrotic damage taken, and the vampire regains hit points equal
      to that amount. The reduction lasts until the target finishes a long
      rest. The target dies if this effect reduces its hit point maximum to
      0. A humanoid slain in this way and then buried in the ground rises
      the following night as a vampire spawn under the vampire's control.
    Charm.
      The vampire targets one humanoid it can see within 30 ft. of it. If
      the target can see the vampire, the target must succeed on a DC 17
      Wisdom saving throw against this magic or be charmed by the vampire.
      The charmed target regards the vampire as a trusted friend to be
      heeded and protected. Although the target isn't under the vampire's
      control, it takes the vampire's requests or actions in the most
      favorable way it can, and it is a willing target for the vampire's bit
      attack.
      
      Each time the vampire or the vampire's companions do anything harmful
      to the target, it can repeat the saving throw, ending the effect on
      itself on a success. Otherwise, the effect lasts 24 hours or until the
      vampire is destroyed, is on a different plane of existence than the
      target, or takes a bonus action to end the effect.
    Children of the Night.
      The vampire magically calls 2d4 swarms of bats or rats, provided that
      the sun isn't up. While outdoors, the vampire can call 3d6 wolves
      instead. The called creatures arrive in 1d4 rounds, acting as allies
      of the vampire and obeying its spoken commands. The beasts remain for
      1 hour, until the vampire dies, or until the vampire dismisses them as
      a bonus action.
    """
    name = "Vampire"
    description = "Medium undead, lawful evil"
    challenge_rating = 13
    armor_class = 16
    skills = "Perception +7, Stealth +9"
    senses = "Darkvision 120 ft., Passive Perception 17"
    languages = "the languages it knew in life"
    strength = Ability(18)
    dexterity = Ability(18)
    constitution = Ability(18)
    intelligence = Ability(17)
    wisdom = Ability(15)
    charisma = Ability(18)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 144
    hit_dice = "17d8"
    spells = []


class VampireSpawn(Monster):
    """
    Regeneration.
      The vampire regains 10 hit points at the start of its turn if it has
      at least 1 hit point and isn't in sunlight or running water. If the
      vampire takes radiant damage or damage from holy water, this trait
      doesn't function at the start of the vampire's next turn.
    Spider Climb.
      The vampire can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Vampire Weaknesses.
      The vampire has the following flaws:
      
      Forbiddance. The vampire can't enter a residence without an invitation
      from one of the occupants.
      
      Harmed by Running Water. The vampire takes 20 acid damage when it ends
      its turn in running water.
      
      Stake to the Heart. The vampire is destroyed if a piercing weapon made
      of wood is driven into its heart while it is incapacitated in its
      resting place.
      
      Sunlight Hypersensitivity. The vampire takes 20 radiant damage when it
      starts its turn in sunlight. While in sunlight, it has disadvantage on
      attack rolls and ability checks.
    Multiattack.
      The vampire makes two attacks, only one of which can be a bite attack.
    Bite.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one willing creature, or
      a creature that is grappled by the vampire, incapacitated, or
      restrained. Hit: 6 (1d6 + 3) piercing damage plus 7 (2d6) necrotic
      damage. The target's hit point maximum is reduced by an amount equal
      to the necrotic damage taken, and the vampire regains hit points equal
      to that amount. The reduction lasts until the target finishes a long
      rest. The target dies if this effect reduces its hit point maximum to
      0.
    Claws.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one creature. Hit: 8 (2d4
      + 3) slashing damage. Instead of dealing damage, the vampire can
      grapple the target (escape DC 13).
    """
    name = "Vampire Spawn"
    description = "Medium undead, neutral evil"
    challenge_rating = 5
    armor_class = 15
    skills = "Perception +3, Stealth +6"
    senses = "Darkvision 60 ft., Passive Perception 13"
    languages = "the languages it knew in life"
    strength = Ability(16)
    dexterity = Ability(16)
    constitution = Ability(16)
    intelligence = Ability(11)
    wisdom = Ability(10)
    charisma = Ability(12)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 82
    hit_dice = "11d8"
    spells = []


class Veteran(Monster):
    """
    Multiattack.
      The veteran makes two longsword attacks. If it has a shortsword drawn,
      it can also make a shortsword attack.
    Longsword.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 7 (1d8 +
      3) slashing damage, or 8 (1d10 + 3) slashing damage if used with two
      hands.
    Shortsword.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 6 (1d6 +
      3) piercing damage.
    Heavy Crossbow.
      Ranged Weapon Attack: +3 to hit, range 100/400 ft., one target. Hit: 6
      (1d10 + 1) piercing damage.
    """
    name = "Veteran"
    description = "Medium humanoid, any alignment"
    challenge_rating = 3
    armor_class = 17
    skills = "Athletics +5, Perception +2"
    senses = "Passive Perception 12"
    languages = "any one language (usually Common)"
    strength = Ability(16)
    dexterity = Ability(13)
    constitution = Ability(14)
    intelligence = Ability(10)
    wisdom = Ability(11)
    charisma = Ability(10)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 58
    hit_dice = "9d8"
    spells = []


class VioletFungus(Monster):
    """
    False Appearance.
      While the violet fungus remains motionless, it is indistinguishable
      from an ordinary fungus.
    Multiattack.
      The fungus makes 1d4 Rotting Touch attacks.
    Rotting Touch.
      Melee Weapon Attack: +2 to hit, reach 10 ft., one creature. Hit: 4
      (1d8) necrotic damage.
    """
    name = "Violet Fungus"
    description = "Medium plant, unaligned"
    challenge_rating = 0.25
    armor_class = 5
    skills = ""
    senses = "Blindsight 30 ft. (blind beyond this radius), Passive Perception 6"
    languages = ""
    strength = Ability(3)
    dexterity = Ability(1)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(3)
    charisma = Ability(1)
    speed = 5
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 18
    hit_dice = "4d8"
    spells = []


class Vrock(Monster):
    """
    Magic Resistance.
      The vrock has advantage on saving throws against spells and other
      magical effects.
    Multiattack.
      The vrock makes two attacks: one with its beak and one with its
      talons.
    Beak.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 10 (2d6
      + 3) piercing damage.
    Talons.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 14 (2d10
      + 3) slashing damage.
    Spores (Recharge 6).
      A 15-foot-radius cloud of toxic spores extends out from the vrock. The
      spores spread around corners. Each creature in that area must succeed
      on a DC 14 Constitution saving throw or become poisoned. While
      poisoned in this way, a target takes 5 (1d10) poison damage at the
      start of each of its turns. A target can repeat the saving throw at
      the end of each of its turns, ending the effect on itself on a
      success. Emptying a vial of holy water on the target also ends the
      effect on it.
    Stunning Screech.
      The vrock emits a horrific screech. Each creature within 20 feet of it
      that can hear it and that isn't a demon must succeed on a DC 14
      Constitution saving throw or be stunned until the end of the vrock's
      next turn .
    Variant: Summon Demon.
      The demon chooses what to summon and attempts a magical summoning.
      
      A vrock has a 30 percent chance of summoning 2d4 dretches or one
      vrock.
      
      A summoned demon appears in an unoccupied space within 60 feet of its
      summoner, acts as an ally of its summoner, and can't summon other
      demons. It remains for 1 minute, until it or its summoner dies, or
      until its summoner dismisses it as an action.
    """
    name = "Vrock"
    description = "Large fiend, chaotic evil"
    challenge_rating = 6
    armor_class = 15
    skills = ""
    senses = "Darkvision 120 ft., Passive Perception 11"
    languages = "Abyssal, telepathy 120 ft."
    strength = Ability(17)
    dexterity = Ability(15)
    constitution = Ability(18)
    intelligence = Ability(8)
    wisdom = Ability(13)
    charisma = Ability(8)
    speed = 40
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 104
    hit_dice = "11d10"
    spells = []


class Vulture(Monster):
    """
    Keen Sight and Smell.
      The vulture has advantage on Wisdom (Perception) checks that rely on
      sight or smell.
    Pack Tactics.
      The vulture has advantage on an attack roll against a creature if at
      least one of the vulture's allies is within 5 ft. of the creature and
      the ally isn't incapacitated.
    Beak.
      Melee Weapon Attack: +2 to hit, reach 5 ft., one target. Hit: 2 (1d4)
      piercing damage.
    """
    name = "Vulture"
    description = "Medium beast, unaligned"
    challenge_rating = 0
    armor_class = 10
    skills = "Perception +3"
    senses = "Passive Perception 13"
    languages = ""
    strength = Ability(7)
    dexterity = Ability(10)
    constitution = Ability(13)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(4)
    speed = 10
    swim_speed = 0
    fly_speed = 50
    climb_speed = 0
    hp_max = 5
    hit_dice = "1d8"
    spells = []



from dungeonsheets.spells.spells import Spell


class RaiseDead(Spell):
    """You return a dead creature you touch to life, provided that it has been dead no
    longer than 10 days. If the creature's soul is both willing and at liberty to
    rejoin the body, the creature returns to life with 1 hit point.

    This spell also
    neutralizes any poison and cures nonmagical diseases that affected the creature
    at the time it died. This spell doesn't, however, remove magical diseases,
    curses, or similar effects; if these aren't first removed prior to casting the
    spell, they take effect when the creature returns to life. The spell can't
    return an undead creature to life.

    This spell closes all mortal wounds, but it
    doesn't restore missing body parts. If the creature is lacking body parts or
    organs integral for its survival – its head, for instance – the spell
    automatically fails.

    Coming back from the dead is an ordeal. The target takes a
    -4 penalty to all attack rolls, saving throws, and ability checks. Every time
    the target finishes a long rest, the penalty is reduced by 1 until it
    disappears.
    """

    name = "Raise Dead"
    level = 5
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A diamond worth at least 500 gp, which the spell consumes"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Bard", "Cleric", "Paladin")


class RarysTelepathicBond(Spell):
    """You forge a telepathic link among up to eight willing creatures of your choice
    within range, psychically linking each creature to all the others for the
    duration. Creatures with Intelligence scores of 2 or less aren't affected by
    this spell.

    Until the spell ends, the targets can communicated telepathically
    through the bond whether or not they have a common language. The communication
    is possible over any distance, though it can't extend to other planes of
    existence.
    """

    name = "Rarys Telepathic Bond"
    level = 5
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "Pieces of eggshell from two different kinds of creatures"
    duration = "1 hour"
    ritual = True
    magic_school = "Divination"
    classes = ("Wizard",)


class RayOfEnfeeblement(Spell):
    """A black beam of enervating energy springs from your finger toward a creature
    within range.
    Make a ranged spell attack against the target. On a hit, the
    target deals only half damage with weapon attacks that use Strength until the
    spell ends.

    At the end of each of the target's turns, it can make a
    Constitution saving throw against the spell. On a success, the spell ends.
    """

    name = "Ray Of Enfeeblement"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Warlock", "Wizard")


class RayOfFrost(Spell):
    """A frigid beam of blue-white light streaks toward a creature within range. Make a
    ranged spell attack against the target. On a hit, it takes 1d8 cold damage, and
    its speed is reduced by 10 feet until the start of your next turn.

    At Higher
    Levels: The spell's damage increases by 1d8 when you reach 5th level (2d8), 11th
    level (3d8), and 17th level (4d8).
    """

    name = "Ray Of Frost"
    level = 0
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class RayOfSickness(Spell):
    """A ray of sickening greenish energy lashes out toward a creature within range.

    Make a ranged spell attack against the target. On a hit, the target takes 2d8
    poison damage and must make a Constitution saving throw. On a failed save, it is
    also poisoned until the end of your next turn.

    **At Higher Levels:** When you cast
    this spell using a spell slot of 2nd level or higher, the damage increases by
    1d8 for each slot level above 1st.
    """

    name = "Ray Of Sickness"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Sorcerer", "Wizard")


class Regenerate(Spell):
    """You touch a creature and stimulate its natural healing ability.
    The target
    regains 4d8 + 15 hit points. For the duration of the spell, the target regains 1
    hit point at the start of each of its turns (10 hit points each minute).

    The
    target's severed body members (fingers, legs, tails, and so on), if any, are
    restored after 2 minutes. If you have the severed part and hold it to the stump,
    the spell instantaneously causes the limb to knit to the stump.
    """

    name = "Regenerate"
    level = 7
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A prayer wheel and holy water"
    duration = "1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Cleric", "Druid")


class Reincarnate(Spell):
    """You touch a dead humanoid or a piece of a dead humanoid. Provided that the
    creature has been dead no longer than 10 days, the spell forms a new adult body
    for it and then calls the soul to enter that body. If the target's soul isn't
    free or willing to do so, the spell fails.

    The magic fashions a new body for
    the creature to inhabit, which likely causes the creature's race to change. The
    DM rolls a d 100 and consults the following table to determine what form the
    creature takes when restored to life, or the DM chooses a form.

    d100  Race

    01-04 Dragonborn
    05-13 Dwarf, hill
    14-21 Dwarf, mountain
    22-25 Elf, dark
    26-34
    Elf, high
    35-42 Elf, wood
    43-46 Gnome, forest
    47-52 Gnome, rock
    53-56  Half-elf

    57-60 Half-orc
    61-68 Halfling, lightfoot
    69-76 Halfling, stout
    77-96 Human
    97-00
    Tiefling

    The reincarnated creature recalls its former life and experiences. It
    retains the capabilities it had in its original form, except it exchanges its
    original race for the new one and changes its racial traits accordingly.
    """

    name = "Reincarnate"
    level = 5
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        """Rare oils and unguents worth at least 1,000 gp, which the spell consumes"""
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid",)


class RemoveCurse(Spell):
    """At your touch, all curses affecting one creature or object end. If the object is
    a cursed magic item, its curse remains, but the spell breaks its owner's
    attunement to the object so it can be removed or discarded.
    """

    name = "Remove Curse"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Paladin", "Warlock", "Wizard")


class Resistance(Spell):
    """You touch one willing creature. Once before the spell ends, the target can roll
    a d4 and add the number rolled to one saving throw of its choice. It can roll
    the die before or after the saving throw. The spell then ends.
    """

    name = "Resistance"
    level = 0
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A miniature cloak"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Druid")


class Resurrection(Spell):
    """You touch a dead creature that has been dead for no more than a century, that
    didn't die of old age, and that isn't undead. If its soul is free and willing,
    the target returns to life with all its hit points.

    This spell neutralizes any
    poisons and cures normal diseases afflicting the creature when it died. It
    doesn't, however, remove magical diseases, curses, and the like; if such affects
    aren't removed prior to casting the spell, they afflict the target on its
    return to life.

    This spell closes all mortal wounds and restores any missing
    body parts.

    Coming back from the dead is an ordeal. The target takes a -4
    penalty to all attack rolls, saving throws, and ability checks. Every time the
    target finishes a long rest, the penalty is reduced by 1 until it disappears.


    Casting this spell to restore life to a creature that has been dead for one year
    or longer taxes you greatly. Until you finish a long rest, you can't cast
    spells again, and you have disadvantage on all attack rolls, ability checks, and
    saving throws.
    """

    name = "Resurrection"
    level = 7
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A diamond worth at least 1,000 gp, which the spell consumes"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Bard", "Cleric")


class ReverseGravity(Spell):
    """This spell reverses gravity in a 50-foot-radius, 100-foot high cylinder centered
    on a point within range.
    All creatures and objects that aren't somehow anchored
    to the ground in the area fall upward and reach the top of the area when you
    cast this spell. A creature can make a Dexterity saving throw to grab onto a
    fixed object it can reach, thus avoiding the fall.

    If some solid object (such
    as a ceiling) is encountered in this fall, falling objects and creatures strike
    it just as they would during a normal downward fall. If an object or creature
    reaches the top of the area without striking anything, it remains there,
    oscillating slightly, for the duration.

    At the end of the duration, affected
    objects and creatures fall back down.
    """

    name = "Reverse Gravity"
    level = 7
    casting_time = "1 action"
    casting_range = "100 feet"
    components = ("V", "S", "M")
    materials = "A lodestone and iron filings"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Wizard")


class Revivify(Spell):
    """You touch a creature that has died within the last minute. That creature returns
    to life with 1 hit point. This spell can't return to life a creature that has
    died of old age, nor can it restore any missing body parts.
    """

    name = "Revivify"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "Diamonds worth 300 gp, which the spell consumes"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Cleric", "Paladin")


class RopeTrick(Spell):
    """You touch a length of rope that is up to 60 feet long. One end of the rope then
    rises into the air until the whole rope hangs perpendicular to the ground. At
    the upper end of the rope, an invisible entrance opens to an extradimensional
    space that lasts until the spell ends.

    The extradimensional space can be
    reached by climbing to the top of the rope. The space can hold as many as eight
    Medium or smaller creatures. The rope can be pulled into the space, making the
    rope disappear from view outside the space.

    Attacks and spells can't cross
    through the entrance into or out of the extradimensional space, but those inside
    can see out of it as if through a 3-foot-by-5-foot window centered on the rope.


    Anything inside the extradimensional space drops out when the spell ends.
    """

    name = "Rope Trick"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "Powdered corn extract and a twisted loop of parchment"
    duration = "1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Wizard",)



import os
import subprocess
import logging
import warnings
import io

import pdfrw
from fdfgen import forge_fdf
from reportlab.pdfgen import canvas

from dungeonsheets.forms import mod_str

CHECKBOX_ON = "Yes"
CHECKBOX_OFF = "Off"
PDFTK_CMD = "pdftk"

log = logging.getLogger(__name__)


def text_box(string):
    """Format a string for displaying in a text box."""
    # remove multiple whitespace without removing linebreaks
    new_string = " ".join(string.replace("\n", "\m").split())  # noqa: W605
    # Remove *single* line breaks, swap *multi* line breaks to single (fdf: \r)
    new_string = (
        new_string.replace("\m \m", "\r")  # noqa: W605
        .replace("\m\m", "\r")  # noqa: W605
        .replace("\m", " ")  # noqa: W605
    )
    return new_string


def create_character_pdf_template(character, basename, flatten=False):
    # Prepare the list of fields
    fields = {
        # Character description
        "CharacterName": character.name,
        "ClassLevel": character.classes_and_levels,
        "Background": str(character.background),
        "PlayerName": character.player_name,
        "Race ": str(character.race),
        "Alignment": character.alignment,
        "XP": str(character.xp),
        "Inspiration": str("Yes" if character.inspiration else ""),
        # Abilities
        "ProfBonus": mod_str(character.proficiency_bonus),
        "STRmod": str(character.strength.value),
        "STR": mod_str(character.strength.modifier),
        "DEXmod ": str(character.dexterity.value),
        "DEX": mod_str(character.dexterity.modifier),
        "CONmod": str(character.constitution.value),
        "CON": mod_str(character.constitution.modifier),
        "INTmod": str(character.intelligence.value),
        "INT": mod_str(character.intelligence.modifier),
        "WISmod": str(character.wisdom.value),
        "WIS": mod_str(character.wisdom.modifier),
        "CHamod": str(character.charisma.value),
        "CHA": mod_str(character.charisma.modifier),
        "AC": str(character.armor_class),
        "Initiative": str(character.initiative),
        "Speed": str(character.speed),
        "Passive": character.passive_wisdom,
        # Saving throws (proficiencies handled later)
        "ST Strength": mod_str(character.strength.saving_throw),
        "ST Dexterity": mod_str(character.dexterity.saving_throw),
        "ST Constitution": mod_str(character.constitution.saving_throw),
        "ST Intelligence": mod_str(character.intelligence.saving_throw),
        "ST Wisdom": mod_str(character.wisdom.saving_throw),
        "ST Charisma": mod_str(character.charisma.saving_throw),
        # Skills (proficiencies handled below)
        "Acrobatics": mod_str(character.acrobatics.modifier),
        "Animal": mod_str(character.animal_handling.modifier),
        "Arcana": mod_str(character.arcana.modifier),
        "Athletics": mod_str(character.athletics.modifier),
        "Deception ": mod_str(character.deception.modifier),
        "History ": mod_str(character.history.modifier),
        "Insight": mod_str(character.insight.modifier),
        "Intimidation": mod_str(character.intimidation.modifier),
        "Investigation ": mod_str(character.investigation.modifier),
        "Medicine": mod_str(character.medicine.modifier),
        "Nature": mod_str(character.nature.modifier),
        "Perception ": mod_str(character.perception.modifier),
        "Performance": mod_str(character.performance.modifier),
        "Persuasion": mod_str(character.persuasion.modifier),
        "Religion": mod_str(character.religion.modifier),
        "SleightofHand": mod_str(character.sleight_of_hand.modifier),
        "Stealth ": mod_str(character.stealth.modifier),
        "Survival": mod_str(character.survival.modifier),
        # Hit points
        "HDTotal": character.hit_dice,
        "HPMax": str(character.hp_max),
        "HPCurrent": str(character.hp_current)
        if character.hp_current is not None
        else "",
        "HPTemp": str(character.hp_temp) if character.hp_temp > 0 else "",
        # Personality traits and other features
        "PersonalityTraits ": text_box(character.personality_traits),
        "Ideals": text_box(character.ideals),
        "Bonds": text_box(character.bonds),
        "Flaws": text_box(character.flaws),
        "Features and Traits": text_box(
            character.features_text + character.features_and_traits
        ),
        # Inventory
        "CP": character.cp,
        "SP": character.sp,
        "EP": character.ep,
        "GP": character.gp,
        "PP": character.pp,
        "Equipment": text_box(character.magic_items_text + character.equipment),
    }
    # Check boxes for proficiencies
    ST_boxes = {
        "strength": "Check Box 11",
        "dexterity": "Check Box 18",
        "constitution": "Check Box 19",
        "intelligence": "Check Box 20",
        "wisdom": "Check Box 21",
        "charisma": "Check Box 22",
    }
    for ability in character.saving_throw_proficiencies:
        fields[ST_boxes[ability]] = CHECKBOX_ON
    # Add skill proficiencies
    skill_boxes = {
        "acrobatics": "Check Box 23",
        "animal_handling": "Check Box 24",
        "arcana": "Check Box 25",
        "athletics": "Check Box 26",
        "deception": "Check Box 27",
        "history": "Check Box 28",
        "insight": "Check Box 29",
        "intimidation": "Check Box 30",
        "investigation": "Check Box 31",
        "medicine": "Check Box 32",
        "nature": "Check Box 33",
        "perception": "Check Box 34",
        "performance": "Check Box 35",
        "persuasion": "Check Box 36",
        "religion": "Check Box 37",
        "sleight_of_hand": "Check Box 38",
        "stealth": "Check Box 39",
        "survival": "Check Box 40",
    }
    for skill in character.skill_proficiencies:
        try:
            fields[skill_boxes[skill.replace(" ", "_").lower()]] = CHECKBOX_ON
        except KeyError:
            raise KeyError(f"Unknown skill: '{skill}'")
    # Add weapons
    weapon_fields = [
        ("Wpn Name", "Wpn1 AtkBonus", "Wpn1 Damage"),
        ("Wpn Name 2", "Wpn2 AtkBonus ", "Wpn2 Damage "),
        ("Wpn Name 3", "Wpn3 AtkBonus  ", "Wpn3 Damage "),
    ]
    for _fields, weapon in zip(weapon_fields, character.weapons):
        name_field, atk_field, dmg_field = _fields
        fields[name_field] = weapon.name
        fields[atk_field] = "{:+d}".format(weapon.attack_modifier)
        fields[dmg_field] = f"{weapon.damage}/{weapon.damage_type}"
    # Additional attacks beyond 3
    attack = [
        f"{w.name}: Atk {w.attack_modifier:+d}, Dam {w.damage}/{w.damage_type}"
        for w in character.weapons[len(weapon_fields) :]
    ]
    # Other attack information
    if character.armor:
        attack.append(f"Armor: {character.armor}")
    if character.shield:
        attack.append(f"Shield: {character.shield}")
    attack.append(character.attacks_and_spellcasting)
    attack_str = "\n\n".join(attack)
    fields["AttacksSpellcasting"] = text_box(attack_str)
    # Other proficiencies and languages
    prof_text = "Proficiencies:\n" + text_box(character.proficiencies_text)
    prof_text += "\n\nLanguages:\n" + text_box(character.languages)
    fields["ProficienciesLang"] = prof_text
    # Prepare the actual PDF
    dirname = os.path.join(os.path.dirname(os.path.abspath(__file__)), "forms/")
    src_pdf = os.path.join(dirname, "blank-character-sheet-default.pdf")
    return make_pdf(fields, src_pdf=src_pdf, basename=basename, flatten=flatten, portrait="")


def create_personality_pdf_template(character, basename, portrait_file="", flatten=False):
    # Prepare the list of fields
    fields = {
        "CharacterName 2": character.name,
        "Age": str(character.age),
        "Height": character.height,
        "Weight": character.weight,
        "Eyes": character.eyes,
        "Skin": character.skin,
        "Hair": character.hair,
        # "CHARACTER IMAGE": None
        # "Faction Symbol Image": None
        "Allies": text_box(character.allies),
        "FactionName": character.faction_name,
        "Backstory": text_box(character.backstory),
        "Feat+Traits": text_box(character.other_feats_traits),
        "Treasure": text_box(character.treasure),
    }
    # Prepare the actual PDF
    dirname = os.path.join(os.path.dirname(os.path.abspath(__file__)), "forms/")
    src_pdf = os.path.join(dirname, "blank-personality-sheet-default.pdf")
    return make_pdf(fields, src_pdf=src_pdf, basename=basename, flatten=flatten, portrait=portrait_file)


def create_spells_pdf_template(character, basename, flatten=False):
    classes_and_levels = " / ".join(
        [c.name + " " + str(c.level) for c in character.spellcasting_classes]
    )
    abilities = " / ".join(
        [c.spellcasting_ability.upper()[:3] for c in character.spellcasting_classes]
    )
    DCs = " / ".join(
        [str(character.spell_save_dc(c)) for c in character.spellcasting_classes]
    )
    bonuses = " / ".join(
        [
            mod_str(character.spell_attack_bonus(c))
            for c in character.spellcasting_classes
        ]
    )

    def spell_level(x):
        return x or 0

    fields = {
        "Spellcasting Class 2": classes_and_levels,
        "SpellcastingAbility 2": abilities,
        "SpellSaveDC  2": DCs,
        "SpellAtkBonus 2": bonuses,
        # Number of spell slots
        "SlotsTotal 19": spell_level(character.spell_slots(1)),
        "SlotsTotal 20": spell_level(character.spell_slots(2)),
        "SlotsTotal 21": spell_level(character.spell_slots(3)),
        "SlotsTotal 22": spell_level(character.spell_slots(4)),
        "SlotsTotal 23": spell_level(character.spell_slots(5)),
        "SlotsTotal 24": spell_level(character.spell_slots(6)),
        "SlotsTotal 25": spell_level(character.spell_slots(7)),
        "SlotsTotal 26": spell_level(character.spell_slots(8)),
        "SlotsTotal 27": spell_level(character.spell_slots(9)),
    }
    # Cantrips
    cantrip_fields = (f"Spells 10{i}" for i in (14, 16, 17, 18, 19, 20, 21, 22))
    cantrips = (spl for spl in character.spells if spl.level == 0)
    for spell, field_name in zip(cantrips, cantrip_fields):
        fields[field_name] = str(spell)
    # Spells for each level
    field_numbers = {
        1: (
            1015,
            1023,
            1024,
            1025,
            1026,
            1027,
            1028,
            1029,
            1030,
            1031,
            1032,
            1033,
        ),
        2: (
            1046,
            1034,
            1035,
            1036,
            1037,
            1038,
            1039,
            1040,
            1041,
            1042,
            1043,
            1044,
            1045,
        ),
        3: (
            1048,
            1047,
            1049,
            1050,
            1051,
            1052,
            1053,
            1054,
            1055,
            1056,
            1057,
            1058,
            1059,
        ),
        4: (
            1061,
            1060,
            1062,
            1063,
            1064,
            1065,
            1066,
            1067,
            1068,
            1069,
            1070,
            1071,
            1072,
        ),
        5: (
            1074,
            1073,
            1075,
            1076,
            1077,
            1078,
            1079,
            1080,
            1081,
        ),
        6: (
            1083,
            1082,
            1084,
            1085,
            1086,
            1087,
            1088,
            1089,
            1090,
        ),
        7: (
            1092,
            1091,
            1093,
            1094,
            1095,
            1096,
            1097,
            1098,
            1099,
        ),
        8: (
            10101,
            10100,
            10102,
            10103,
            10104,
            10105,
            10106,
        ),
        9: (10108, 10107, 10109, 101010, 101011, 101012, 101013),
    }
    prep_numbers = {
        1: (
            251,
            309,
            3010,
            3011,
            3012,
            3013,
            3014,
            3015,
            3016,
            3017,
            3018,
            3019,
        ),
        2: (
            313,
            310,
            3020,
            3021,
            3022,
            3023,
            3024,
            3025,
            3026,
            3027,
            3028,
            3029,
            3030,
        ),
        3: (
            315,
            314,
            3031,
            3032,
            3033,
            3034,
            3035,
            3036,
            3037,
            3038,
            3039,
            3040,
            3041,
        ),
        4: (
            317,
            316,
            3042,
            3043,
            3044,
            3045,
            3046,
            3047,
            3048,
            3049,
            3050,
            3051,
            3052,
        ),
        5: (
            319,
            318,
            3053,
            3054,
            3055,
            3056,
            3057,
            3058,
            3059,
        ),
        6: (
            321,
            320,
            3060,
            3061,
            3062,
            3063,
            3064,
            3065,
            3066,
        ),
        7: (
            323,
            322,
            3067,
            3068,
            3069,
            3070,
            3071,
            3072,
            3073,
        ),
        8: (
            325,
            324,
            3074,
            3075,
            3076,
            3077,
            3078,
        ),
        9: (
            327,
            326,
            3079,
            3080,
            3081,
            3082,
            3083,
        ),
    }
    # Prepare the lists of spells for each level
    for level in field_numbers.keys():
        spells = [spl for spl in character.spells if spl.level == level]
        # Determine if we should omit un-prepared spells to save space
        if len(spells) > len(field_numbers[level]):
            spells = [s for s in spells if s in character.spells_prepared]
            warnings.warn(
                f"{character.name} knows more spells than the number of "
                "lines available in spell sheet. Limited to prepared "
                "spells only."
            )
        # Build the list of PDF controls to set/toggle
        field_names = [f"Spells {i}" for i in field_numbers[level]]
        prep_names = tuple(f"Check Box {i}" for i in prep_numbers[level])
        for spell, field, chk_field in zip(spells, field_names, prep_names):
            fields[field] = str(spell)
            is_prepared = any([spell == Spl for Spl in character.spells_prepared])
            fields[chk_field] = CHECKBOX_ON if is_prepared else CHECKBOX_OFF
        # # Uncomment to post field names instead:
        # for field in field_names:
        #     fields.append((field, field))
    # Make the actual pdf
    dirname = os.path.join(os.path.dirname(os.path.abspath(__file__)), "forms/")
    src_pdf = os.path.join(dirname, "blank-spell-sheet-default.pdf")
    make_pdf(fields, src_pdf=src_pdf, basename=basename, flatten=flatten, portrait="")


def make_pdf(fields: dict, src_pdf: str, basename: str, flatten: bool = False, portrait = ""):
    """Create a new PDF by applying fields to a src PDF document.

    Parameters
    ==========
    fields :
      Data to fill into the form. The keys are field names in the PDF
      form, and the values will be entered as the value in the PDF.
    src_pdf :
      Path to the PDF that will serve as the template.
    basename :
      The path of the destination PDF without the file extensions. The
      resulting pdf will be {basename}.pdf
    flatten :
      If truthy, the PDF will be collapsed so it is no longer
      editable.

    """
    try:
        _make_pdf_pdftk(fields, src_pdf, basename, flatten, portrait)
    except FileNotFoundError:
        # pdftk could not run, so alert the user and use pdfrw
        warnings.warn(
            f"Could not run `{PDFTK_CMD}`, using fallback; forcing `--editable`.",
            RuntimeWarning,
        )
        _make_pdf_pdfrw(fields, src_pdf, basename, flatten, portrait)


def _make_pdf_pdfrw(fields: dict, src_pdf: str, basename: str, flatten: bool = False, portrait = ""):
    """Backup make_pdf function in case pdftk is not available."""
    template = pdfrw.PdfReader(src_pdf)
    # Different types of PDF fields
    BUTTON = "/Btn"
    # Names for entries in PDF annotation list
    # DEFAULT_VALUE = "/DV"
    # APPEARANCE = "/MK"
    FIELD = "/T"
    # PROPS = "/P"
    TYPE = "/FT"
    # FLAGS = "/Ff"
    # SUBTYPE = "/Subtype"
    # ALL_KEYS = [
    #     "/DV",
    #     "/F",
    #     "/FT",
    #     "/Ff",
    #     "/MK",
    #     "/P",
    #     "/Rect",
    #     "/Subtype",
    #     "/T",
    #     "/Type",
    # ]
    annots = template.pages[0]["/Annots"]
    # Update each annotation if it's in the requested dictionary
    for annot in annots:
        this_field = annot[FIELD][1:-1]
        # Check if the field has a new value passed
        if this_field in fields.keys():
            val = fields[this_field]
            # Convert integers to strings
            if isinstance(val, int):
                val = str(val)
            log.debug(
                f"Set field '{this_field}' "
                f"({annot[TYPE]}) "
                f"to `{val}` ({val.__class__}) "
                f"in file '{basename}.pdf'"
            )
            # Prepare a PDF dictionary based on the fields properties
            if annot[TYPE] == BUTTON:
                # Radio buttons require special appearance streams
                if val == CHECKBOX_ON:
                    val = bytes(val, "utf-8")
                    pdf_dict = pdfrw.PdfDict(V=val, AS=val)
                else:
                    continue
            else:
                # All other widget types
                pdf_dict = pdfrw.PdfDict(V=val)
            annot.update(pdf_dict)
        else:
            log.debug(f"Skipping unused field '{this_field}' in file '{basename}.pdf'")
    # Now write the PDF to the new pdf file
    pdfrw.PdfWriter().write(f"{basename}.pdf", template)


def _make_pdf_pdftk(fields, src_pdf, basename, flatten=False, portrait=""):
    """More robust way to make a PDF, but has a hard dependency."""
    # Create the actual FDF file
    fdfname = basename + ".fdf"

    fdf = forge_fdf("", fields, [], [], [])
    fdf_file = open(fdfname, "wb")
    fdf_file.write(fdf)
    fdf_file.close()
    # Build the final flattened PDF documents
    if portrait != "":
        dest_pdf = basename + "-temp.pdf"
        image_pdf = basename + "_image_tmp.pdf"
        make_image_pdf(portrait, image_pdf)
    else:
        dest_pdf = basename + ".pdf"
    popenargs = [
        PDFTK_CMD,
        src_pdf,
        "fill_form",
        fdfname,
        "output",
        dest_pdf,
    ]
    if flatten:
        popenargs.append("flatten")
    subprocess.call(popenargs)
    # stamp with image
    if portrait != "":
        src_pdf = dest_pdf
        stamped_pdf = basename + ".pdf"
        popenargs = [
            PDFTK_CMD,
            src_pdf,
            "stamp",
            image_pdf,
            "output",
            stamped_pdf,
        ]
        popenargs.append("flatten")
        subprocess.call(popenargs)
        # Clean up
        os.remove(image_pdf)
        os.remove(dest_pdf)
    # Clean up temporary files
    os.remove(fdfname)

def make_image_pdf(src_img:str, dest_pdf:str):
    packet = io.BytesIO()
    can = canvas.Canvas(packet)
    x_start = 10
    y_start = 240
    can.drawImage(src_img, x_start, y_start, width=175, preserveAspectRatio=True, mask='auto')
    can.showPage()
    can.save()
 
    #move to the beginning of the StringIO buffer
    packet.seek(0)
 
    new_pdf = pdfrw.PdfReader(packet)
    pdfrw.PdfWriter().write(dest_pdf, new_pdf)
